#
# Copyright (c) 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import re
import yaml
import deepdiff
import logging

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO
from pecan import conf

from managesf.model.yamlbkd.yamlbackend import YAMLBackend
from managesf.model.yamlbkd.yamlbackend import MemoryYAMLBackend
from managesf.model.yamlbkd.yamlbackend import YAMLDBException
from managesf.model.yamlbkd.resource import ModelInvalidException
from managesf.model.yamlbkd.resource import ResourceInvalidException
from managesf.model.yamlbkd.resource import KEY_RE_CONSTRAINT

from managesf.model.yamlbkd.resources.gitrepository import GitRepository
from managesf.model.yamlbkd.resources.group import Group
from managesf.model.yamlbkd.resources.gitacls import ACL
from managesf.model.yamlbkd.resources.project import Project
from managesf.model.yamlbkd.resources.tenant import Tenant
from managesf.model.yamlbkd.resources.connection import Connection


logger = logging.getLogger(__name__)

MAPPING = {'repos': GitRepository,
           'groups': Group,
           'projects': Project,
           'acls': ACL,
           'tenants': Tenant,
           'connections': Connection}


class ResourceDepsException(Exception):
    pass


class ResourceUnicityException(Exception):
    pass


class RTYPENotDefinedException(Exception):
    pass


class SFResourceBackendEngine(object):
    def __init__(self, workdir, subdir):
        self.workdir = workdir
        self.subdir = subdir
        logger.info('Resource engine is using %s as workdir' % (
                    self.workdir))

    def _get_update_change(self, rtype, prev, new, rids):
        """ Based on the resources id known to have evolved
        (was updated) we use the DeepDiff library identify
        the resource keys where values have changed between
        the prev[rid] and new[rid] struct. DeepDiff permits
        to easily known if a tuple/list contains changes as
        well as simple type like str.

        This function acts at the resource[rtype][rid]
        level only.
        """
        r_key_changes = {}
        for rid in rids:
            n = MAPPING[rtype](rid, new[rid])
            p = MAPPING[rtype](rid, prev[rid])
            n.set_defaults()
            p.set_defaults()
            new_data = n.get_resource()
            prev_data = p.get_resource()
            changes = deepdiff.DeepDiff(prev_data, new_data)
            if not changes:
                continue
            r_key_changes[rid] = {'data': new_data, 'changed': set()}
            for ctype, _changes in changes.items():
                if ctype in ('values_changed', 'iterable_item_removed',
                             'iterable_item_added', 'type_changes',
                             'dictionary_item_added',
                             'dictionary_item_removed'):
                    for c in _changes:
                        key = re.search(
                            r"root\['(" + KEY_RE_CONSTRAINT + r")'\].*$", c)
                        key = key.groups()[0]
                        r_key_changes[rid]['changed'].add(key)
                else:
                    logger.info('Unexpected change type (%s) '
                                'detected for rid: %s' % (ctype, rid))
        return r_key_changes

    def _get_data_diff(self, prev, new):
        """ Top level resources diff method that take two
        resources trees and compute resources added/removed
        and updated.

        This method returns a dict containing ids and data
        of resources by resource type and modification type
        create, delete, update.
        """
        previous_data_resources_ids = {}
        for rtype, resources in prev['resources'].items():
            previous_data_resources_ids[rtype] = set(resources.keys())
        new_data_resources_ids = {}
        for rtype, resources in new['resources'].items():
            new_data_resources_ids[rtype] = set(resources.keys())
        changed_resources_ids = {}
        new_resources_ids = {}
        removed_resources_ids = {}
        rtype_list = (set(previous_data_resources_ids.keys()) |
                      set(new_data_resources_ids.keys()))
        for rtype in rtype_list:
            changed_resources_ids[rtype] = (
                previous_data_resources_ids[rtype] &
                new_data_resources_ids[rtype])
            new_resources_ids[rtype] = (
                new_data_resources_ids[rtype] -
                previous_data_resources_ids[rtype])
            removed_resources_ids[rtype] = (
                previous_data_resources_ids[rtype] -
                new_data_resources_ids[rtype])
        sanitized_changes = {}
        for rtype in rtype_list:
            sanitized_changes.setdefault(
                rtype,
                {'update': {}, 'create': {}, 'delete': {}})
            sanitized_changes[rtype]['create'].update(dict([
                (rid, d) for rid, d in new['resources'][rtype].items()
                if rid in new_resources_ids[rtype]]))
            sanitized_changes[rtype]['delete'].update(dict([
                (rid, d) for rid, d in prev['resources'][rtype].items()
                if rid in removed_resources_ids[rtype]]))
            sanitized_changes[rtype]['update'] = self._get_update_change(
                rtype, prev['resources'][rtype], new['resources'][rtype],
                changed_resources_ids[rtype])
        return sanitized_changes

    def _check_rtype_defined(self, data):
        for rtype in data['resources']:
            if rtype not in MAPPING:
                raise RTYPENotDefinedException(
                    "Resources type: %s not exists" % rtype)

    def _check_unicity_constraints(self, new_data):
        """ This method read the new tree and validate if each
        each resources PRIMARY_KEY are unique over the same
        set of type of resources
        """
        for rtype, resources in new_data['resources'].items():
            accu = {}
            pk = MAPPING[rtype].PRIMARY_KEY
            if not pk:
                continue
            for rid, data in resources.items():
                if data[pk] in accu:
                    raise ResourceUnicityException(
                        "Resource [type: %s, ID: %s] primary key (%s) "
                        "%s is already used by another resource" % (
                            rtype, rid, pk, data[pk]))
                else:
                    accu[data[pk]] = None

    def _check_deps_constraints(self, new_data):
        """ This method read the new tree and validate if each
        id returned by each resources get_deps method is found
        in the new tree.
        """
        for rtype, resources in new_data['resources'].items():
            for rid, data in resources.items():
                r = MAPPING[rtype](rid, data)
                if r.is_deps_soft():
                    # This resource do not require deps existence check
                    continue
                r.set_defaults()
                deps = r.get_deps()
                for deps_type, deps_ids in deps.items():
                    for deps_id in deps_ids:
                        try:
                            assert deps_type in new_data['resources']
                            assert deps_id in tuple(
                                new_data['resources'][deps_type])
                        except AssertionError:
                            raise ResourceDepsException(
                                "Resource [type: %s, ID: %s] depends on an "
                                "unknown resource [type: %s, ID: %s]" % (
                                    rtype, rid, deps_type, deps_id))

    def _validate_changes(self, sanitized_changes, validation_logs, new):
        """ This method validates if changes on the tree is authorized
        based on each resources constraints.
        """
        for rtype, changes in sanitized_changes.items():
            for ctype, scoped_changes in changes.items():
                if ctype == 'create':
                    for rid, data in scoped_changes.items():
                        # Full new resource validation
                        r = MAPPING[rtype](rid, data)
                        r.validate()
                        r.set_defaults()
                        xv = MAPPING[rtype].CALLBACKS['extra_validations']
                        logs = xv(conf, new, r.get_resource())
                        if logs:
                            validation_logs.extend(logs)
                            raise ResourceInvalidException(
                                "Resource [type: %s, ID: %s] extra "
                                "validations failed" % (rtype, rid))
                        validation_logs.append(
                            "Resource [type: %s, ID: %s] is going to "
                            "be created." % (rtype, rid))
                if ctype == 'update':
                    for rid, data in scoped_changes.items():
                        # Full new resource validation
                        r = MAPPING[rtype](rid, data['data'])
                        r.validate()
                        xv = MAPPING[rtype].CALLBACKS['extra_validations']
                        logs = xv(conf, new, data['data'])
                        if logs:
                            validation_logs.extend(logs)
                            raise ResourceInvalidException(
                                "Resource [type: %s, ID: %s] extra "
                                "validations failed" % (rtype, rid))
                        # Check key changes are possible
                        if not all([r.is_mutable(k) for
                                    k in data['changed']]):
                            raise YAMLDBException(
                                "Resource [type: %s, ID: %s] contains changed "
                                "resource keys that are immutable. "
                                "Please check the model." % (
                                    rtype, rid))
                        validation_logs.append(
                            "Resource [type: %s, ID: %s] is going to "
                            "be updated." % (rtype, rid))
                if ctype == 'delete':
                    for rid, data in scoped_changes.items():
                        validation_logs.append(
                            "Resource [type: %s, ID: %s] is going to "
                            "be deleted." % (rtype, rid))

    def _get_resources_priority(self):
        """ This method determines in which order resources changes
        must be applied on the services. It is based on the PRIORITY
        values of each resource types.
        """
        priorities = [(rtype, obj.PRIORITY) for
                      rtype, obj in MAPPING.items()]
        return sorted(priorities, key=lambda p: p[1], reverse=True)

    def _apply_changes(self, sanitized_changes, apply_logs, new):
        """ This method apply detected changes to services via
        the resources callbacks and add to a logs set successful
        and failed ops. Callbacks are called in the right order
        related to the PRIORITY.
        """
        partial_errors = False
        priorities = self._get_resources_priority()
        for rtype, priority in priorities:
            if rtype not in sanitized_changes:
                continue
            for ctype, datas in sanitized_changes[rtype].items():
                for rid, data in datas.items():
                    apply_logs.append(
                        "Resource [type: %s, ID: %s] will be %s." % (
                            rtype, rid, ctype + 'd'))
                    logs = []
                    try:
                        if ctype == 'update':
                            # Resource set_defaults is done in
                            # get_update_change for resources that
                            # need to be updated
                            logs = MAPPING[rtype].CALLBACKS[ctype](
                                conf, new, data['data'])
                        else:
                            r = MAPPING[rtype](rid, data)
                            r.set_defaults()
                            _data = r.get_resource()
                            logs = MAPPING[rtype].CALLBACKS[ctype](
                                conf, new, _data)
                    except Exception as e:
                        logger.exception("apply_changes failed %s" % e)
                        logs.append(
                            "Resource [type: %s, ID: %s] %s op error (%s)." % (
                                rtype, rid, ctype, str(e)))
                    if logs:
                        # We have logs here meaning callback call
                        # has encountered some troubles
                        apply_logs.extend(logs)
                        partial_errors = True
                        apply_logs.append(
                            "Resource [type: %s, ID: %s] %s op failed." % (
                                rtype, rid, ctype))
                    else:
                        apply_logs.append(
                            "Resource [type: %s, ID: %s] has been %s." % (
                                rtype, rid, ctype + 'd'))
        return partial_errors

    def _resolv_resources_need_refresh(self, sanitized_changes, tree):
        """ This method detects which resources need to be updated
        because resources it depends on have been updated. For instance
        if an ACLs has be updated then the project that depends on it
        need to be updated.
        """
        def scan(logs):
            u_resources = {}
            for rtype in sanitized_changes:
                u_resources[rtype] = sanitized_changes[rtype]['update'].keys()

            if not any([len(ids) for ids in u_resources.values()]):
                return logs

            for rtype, resources in tree['resources'].items():
                for rid, data in resources.items():
                    r = MAPPING[rtype](rid, data)
                    if not r.should_be_updated():
                        # This resource does not need to be refreshed
                        # as it depends on a resource type that an update on it
                        # has no impact on this resource.
                        # eg. An ACL type depends on group type but a group
                        # update does not need to trigger an ACL update.
                        continue
                    deps = r.get_deps()
                    for deps_type, deps_ids in deps.items():
                        if (deps_type in u_resources and
                                deps_ids & set(u_resources[deps_type])):
                            sanitized_changes.setdefault(rtype, {'update': {}})
                            if rid not in sanitized_changes[rtype]['update']:
                                logs.append(
                                    "Resource [type: %s, ID: %s] need a "
                                    "refresh as at least one of its "
                                    "dependencies has been updated" % (
                                        rtype, rid))
                                sanitized_changes[rtype]['update'][rid] = {
                                    'data': data}
        logs = []
        # Do it 3 times to resolv deps of max 3 depths
        for _ in range(3):
            scan(logs)
        return logs

    def _append_missing_rtype(self, data):
        # If a tree leaf is missing for rtype then add
        # it by default. More convenient to avoid more
        # check later in the code.
        data.setdefault('resources', {})
        for rtype in MAPPING:
            data['resources'].setdefault(rtype, {})

    def _load_resource_data(self, repo_uri, ref, mark):
        """ Read the tree from YAML files stored in
        a GIT repository at a specific ref.
        """
        cpath = os.path.join(self.workdir, mark)
        if not os.path.isdir(cpath):
            os.mkdir(cpath)

        bkd = YAMLBackend(repo_uri, ref,
                          self.subdir, cpath,
                          "%s_cache" % cpath)

        data = bkd.get_data()
        self._append_missing_rtype(data)
        return data

    def _load_resource_data_from_memory(self, data):
        data = MemoryYAMLBackend(data).get_data()
        self._append_missing_rtype(data)
        return data

    def _load_resources_data(self, repo_prev_uri, prev_ref,
                             repo_new_uri, new_ref):
        """ Load data from two GIT repo refs
        """
        prev_data = self._load_resource_data(
            repo_prev_uri, prev_ref, 'prev')
        new_data = self._load_resource_data(
            repo_new_uri, new_ref, 'new')
        return prev_data, new_data

    def _get_missing_resources_diff(self, current, reality, ret_tree):
        """ This method compute the resources read from the
        reality that miss the current (suppose to be the
        config/resources/ tree master) tree. The computing
        is based on the resources PRIMARY_KEY.

        Resources read from the reality will be attached
        to an auto computed rid. These rid will surely differ
        from what users will use when declaring a resource.

        Looking at each resources PRIMARY_KEY value let us
        detect of resource that exist in the current tree
        under a different rid.

        Furthermore dependency ids of resources from reality
        will be updated in the proposed tree (ret_tree) to
        match the current ones.

        Finally the method fill ret_tree with only missing
        resources.
        """
        current_cache_ids = {}
        already_exist_ids = {}
        # Create a cache by resource type with dict
        # {'PRIMARY_KEY_VALUE': rid} - This to speedup lookup.
        for rtype, resources in current['resources'].items():
            current_cache_ids.setdefault(rtype, {})
            for rid, data in resources.items():
                pk = MAPPING[rtype].PRIMARY_KEY
                r = MAPPING[rtype](rid, data)
                current_cache_ids[rtype][r.resource[pk]] = rid

        # Fill the ret_tree with non existing resources
        # compared to current and keep track of already
        # existing one ids.
        for rtype, resources in reality['resources'].items():
            for rid, data in resources.items():
                pk = MAPPING[rtype].PRIMARY_KEY
                r = MAPPING[rtype](rid, data)
                m_rid = current_cache_ids[rtype].get(
                    r.resource[pk], None)
                if not m_rid:
                    # The resource does not exists in the
                    # current tree - so keep it
                    ret_tree['resources'].setdefault(rtype, {})
                    ret_tree['resources'][rtype][rid] = data
                else:
                    # The resource already exists in the
                    # current tree - mark it as existing
                    already_exist_ids.setdefault(rtype, {})
                    already_exist_ids[rtype][rid] = m_rid

        # Finally update dependencies ids if needed
        for rtype, resources in ret_tree['resources'].items():
            for rid, data in resources.items():
                r = MAPPING[rtype](rid, data)
                dkey = r.get_deps(keyname=True)
                if not dkey:
                    continue
                rdeps = r.get_deps()
                resolved_rdeps = []
                drtype = list(rdeps)[0]
                dids = rdeps[drtype]
                if dids:
                    for did in dids:
                        if (drtype in already_exist_ids and
                           did in already_exist_ids[drtype]):
                            resolved_rdeps.append(
                                already_exist_ids[drtype][did])
                        else:
                            resolved_rdeps.append(did)
                    if r.MODEL[dkey][0] is list:
                        r.resource[dkey] = resolved_rdeps
                    elif r.MODEL[dkey][0] is str:
                        r.resource[dkey] = resolved_rdeps[0]

        # Remove the name key as the rid is the name
        for rtype, resources in ret_tree['resources'].items():
            for rid, data in resources.items():
                del ret_tree['resources'][rtype][rid]['name']

    def validate(self, repo_prev_uri, prev_ref,
                 repo_new_uri, new_ref):
        """ Top level validate function
        """
        logger.info("Resources engine: validate resources requested "
                    "(old ref: %s, new ref: %s)" % (prev_ref, new_ref))
        if not os.path.isdir(self.workdir):
            os.mkdir(self.workdir)
        validation_logs = []
        try:
            prev, new = self._load_resources_data(
                repo_prev_uri, prev_ref, repo_new_uri, new_ref)
            self._check_rtype_defined(new)
            self._check_deps_constraints(new)
            self._check_unicity_constraints(new)
            changes = self._get_data_diff(prev, new)
            self._validate_changes(changes, validation_logs, new)
            validation_logs.extend(
                self._resolv_resources_need_refresh(changes, new))
        except (YAMLDBException,
                RTYPENotDefinedException,
                ModelInvalidException,
                ResourceInvalidException,
                ResourceUnicityException,
                ResourceDepsException) as e:
            validation_logs.append(str(e))
            for line in validation_logs:
                logger.info(line)
            return False, validation_logs
        for line in validation_logs:
            logger.info(line)
        return True, validation_logs

    def validate_from_structured_data(self, repo_prev_uri, prev_ref, data):
        """ Top level validate_from_structured_data function
        """
        logger.info("Resources engine: validate_from_structured_data resources"
                    " requested (old ref: %s, structured_data)" % prev_ref)
        if not os.path.isdir(self.workdir):
            os.mkdir(self.workdir)
        validation_logs = []
        try:
            prev = self._load_resource_data(
                repo_prev_uri, prev_ref, 'prev')
            new = self._load_resource_data_from_memory(data)
            self._check_rtype_defined(new)
            self._check_deps_constraints(new)
            self._check_unicity_constraints(new)
            changes = self._get_data_diff(prev, new)
            self._validate_changes(changes, validation_logs, new)
            validation_logs.extend(
                self._resolv_resources_need_refresh(changes, new))
        except (YAMLDBException,
                RTYPENotDefinedException,
                ModelInvalidException,
                ResourceInvalidException,
                ResourceUnicityException,
                ResourceDepsException) as e:
            validation_logs.append(str(e))
            for line in validation_logs:
                logger.info(line)
            return False, validation_logs
        for line in validation_logs:
            logger.info(line)
        return True, validation_logs

    def apply(self, repo_prev_uri, prev_ref,
              repo_new_uri, new_ref):
        """ Top level apply function
        """
        logger.info("Resources engine: apply resources requested"
                    "(old ref: %s, new ref: %s)" % (prev_ref, new_ref))
        if not os.path.isdir(self.workdir):
            os.mkdir(self.workdir)
        apply_logs = []
        try:
            prev, new = self._load_resources_data(
                repo_prev_uri, prev_ref, repo_new_uri, new_ref)
            changes = self._get_data_diff(prev, new)
            logs = self._resolv_resources_need_refresh(changes, new)
            apply_logs.extend(logs)
            partial = self._apply_changes(changes, apply_logs, new)
        except YAMLDBException as e:
            apply_logs.append(str(e))
            for line in apply_logs:
                logger.info(line)
            return False, apply_logs
        for line in apply_logs:
            logger.info(line)
        return not partial, apply_logs

    def get(self, cur_uri, cur_ref, public_url=None):
        """ Top level get function. This read the HEAD of the
        repo and return the resources data tree unmodified.
        """
        logger.info("Resources engine: get resource tree requested")
        if not os.path.isdir(self.workdir):
            os.mkdir(self.workdir)
        current = YAMLBackend(cur_uri, cur_ref,
                              self.subdir, self.workdir,
                              "%s_cache" % self.workdir.rstrip('/'))
        data = current.get_data()
        data_trans = {"resources": {}}
        for rtype, resources in data.get('resources', {}).items():
            for rid, rdata in resources.items():
                r = MAPPING[rtype](rid, rdata)
                r.set_defaults(soft=True)
                dtrans = r.transform_for_get()
                data_trans["resources"].setdefault(rtype, {})[rid] = dtrans
        data_trans["config-repo"] = cur_uri
        data_trans["public-url"] = (
            public_url or conf.resources.get("public_url"))
        # If we have connections defined in the configuration than mean
        # we are on a tenant deployement. So can set the connections resources
        # key from the config
        if hasattr(
                conf, 'resources') and conf.resources.get('connections', {}):
            data_trans['resources'][
                'connections'] = conf.resources.get('connections', {})
        return data_trans

    def direct_apply(self, prev, new):
        """ Top level direct_apply function. This function should be
        called only under specific conditions.

        The yamls will be checked for consistencies then resources
        deduced from the diff of both will be applied. It is needed to
        understand that using this will de-synchronize the config
        respository from the reality.
        """
        logger.info("Resources engine: direct apply resources requested")
        direct_apply_logs = []
        try:
            try:
                prev = yaml.safe_load(StringIO(prev))
            except Exception:
                raise YAMLDBException("YAML format corrupted for prev")
            try:
                new = yaml.safe_load(StringIO(new))
            except Exception:
                raise YAMLDBException("YAML format corrupted for new")
            YAMLBackend._validate_base_struct(prev)
            YAMLBackend._validate_base_struct(new)
            for rtype in MAPPING:
                prev['resources'].setdefault(rtype, {})
                new['resources'].setdefault(rtype, {})
            self._check_rtype_defined(prev)
            self._check_rtype_defined(new)
            self._check_deps_constraints(new)
            self._check_unicity_constraints(new)
            changes = self._get_data_diff(prev, new)
            self._validate_changes(changes, direct_apply_logs, new)
            direct_apply_logs.extend(
                self._resolv_resources_need_refresh(changes, new))
            partial = self._apply_changes(changes, direct_apply_logs, new)
        except (YAMLDBException,
                RTYPENotDefinedException,
                ModelInvalidException,
                ResourceInvalidException,
                ResourceUnicityException,
                ResourceDepsException) as e:
            direct_apply_logs.append(str(e))
            for line in direct_apply_logs:
                logger.info(line)
            return False, direct_apply_logs
        for line in direct_apply_logs:
            logger.info(line)
        if partial:
            return False, direct_apply_logs
        return True, direct_apply_logs

    def get_missing_resources(self, cur_uri, cur_ref):
        """ Top level get_missing_resources. This method read
        the real resources from services and return resources
        struct containing only missing resources (not found
        in the YAML backend (config repo master HEAD) but on
        services)
        """
        logger.info("Resources engine: get missing resources diff requested")
        current = self.get(cur_uri, cur_ref)
        reality = {'resources': {}}
        for rtype in MAPPING:
            current['resources'].setdefault(rtype, {})
            reality['resources'].setdefault(rtype, {})
        ret_tree = {'resources': {}}
        logs = []
        for rtype in MAPPING:
            ret = MAPPING[rtype].CALLBACKS['get_all'](conf, {})
            logs.extend(ret[0])
            reality['resources'].update(ret[1])
        # Get diff between reality and config
        self._get_missing_resources_diff(current, reality, ret_tree)
        for line in logs:
            logger.info(line)
        return logs, ret_tree



#
# Copyright (C) 2018 Red Hat
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import os
import sys
import json
import shutil
import logging
import argparse
import tempfile
import requests
import subprocess

from contextlib import contextmanager

from pecan import conf
from pecan import configuration

from managesf.lib.common import get_managesf_info
from managesf.lib.auth import get_auth_engine

from managesf.model.yamlbkd.engine import SFResourceBackendEngine


@contextmanager
def TemporaryDirectory():
    name = tempfile.mkdtemp()
    try:
        yield name
    finally:
        shutil.rmtree(name)


def read_data_to_validate():
    data = dict(
        [(os.path.join('resources', f),
          open(os.path.join('resources', f)).read())
         for f in os.listdir('resources')
            if f.endswith('.yaml') or f.endswith('.yml')])
    return data


def read_repo_to_validate(zuul_prev_commit, zuul_commit):
    if not os.path.isdir('.git') or not os.path.isdir('resources'):
        print("Current workding directory must the config repository "
              "to validate")
        sys.exit(1)
    if zuul_commit:
        subprocess.call(["git", "checkout", zuul_commit])
    else:
        zuul_commit = ''
    if zuul_prev_commit == 'master':
        zuul_prev_commit = 'origin/master'
    commit_range = "%s..%s" % (zuul_prev_commit, zuul_commit)
    head_commit_msg = subprocess.check_output(
        ["git", "log", commit_range]).decode()
    print(head_commit_msg.encode('ascii', errors='ignore').decode('ascii'))
    structured_data = read_data_to_validate()
    return structured_data, head_commit_msg


def is_commit_exists(commit):
    return subprocess.Popen(["git", "show", "--quiet", commit]).wait() == 0


def display_warnings(logs, head_commit_msg):
    if "is going to be deleted" in " ".join(logs):
        if "sf-resources: allow-delete" not in head_commit_msg:
            print(
                "\n"
                "Resources deletion(s) have been detected.\n"
                "The commit msg tag: 'sf-resources: allow-delete' has not "
                "been detected.\n"
                "The change won't be validated until you include the "
                "tag 'sf-resources: allow-delete' in the commit message.")
            sys.exit(1)
    if "sf-resources: skip-apply" in " ".join(logs):
        print(
            "\n"
            "The commit msg tag: 'sf-resources: skip-apply' has been "
            "detected.\n"
            "The approval of this patch won't trigger the creation of "
            "the resources above.\n"
            "The purpose of this commit is usually to re-sync "
            "config/resources with the reality.")


def cli():
    p = argparse.ArgumentParser()
    p.add_argument(
        "--cache-dir", default="/var/lib/software-factory/managesf-resources")
    p.add_argument(
        "--managesf-config", default="/etc/managesf/config.py")
    p.add_argument(
        "--zuul-commit",
        help="Reference of the commit to apply")
    p.add_argument(
        "--zuul-prev-commit",
        help="Reference of the previous commit applied")
    p.add_argument(
        "--prev-yaml",
        help="Path to the previous yaml (direct-apply)")
    p.add_argument(
        "--new-yaml",
        help="Path to the new yaml (direct-apply)")
    p.add_argument(
        "--log-output", help="Engine logs output file path")
    p.add_argument(
        "--remote-gateway", help="URL to the remote SF")
    p.add_argument("--debug", action="store_true")
    p.add_argument(
        "action", choices=[
            "read", "validate", "remote-validate", "apply", "direct-apply"])
    args = p.parse_args()

    kwargs = {}
    if args.log_output:
        kwargs.update({'filename': args.log_output})

    logging.basicConfig(
        format='%(asctime)s %(levelname)-5.5s %(name)s - %(message)s',
        level=logging.DEBUG if args.debug else logging.INFO,
        **kwargs)

    if not os.path.isdir(os.path.expanduser(args.cache_dir)):
        os.makedirs(os.path.expanduser(args.cache_dir))

    if args.action not in ['remote-validate']:
        configuration.set_config(args.managesf_config)
        engine = SFResourceBackendEngine(
            os.path.join(args.cache_dir, args.action),
            conf.resources.subdir)

    if args.action == "read":
        if not args.zuul_commit:
            args.zuul_commit = 'master'
        raw = engine.get(
            conf.resources.master_repo, args.zuul_commit)
        print(json.dumps(raw, indent=2, sort_keys=True))

    if args.action == "validate":
        if not args.zuul_prev_commit:
            args.zuul_prev_commit = "master"
        structured_data, head_commit_msg = read_repo_to_validate(
            args.zuul_prev_commit, args.zuul_commit)
        status, logs = engine.validate_from_structured_data(
            conf.resources.master_repo, args.zuul_prev_commit, structured_data)
        print("")
        print(
            "=== Resources actions list that will apply once patch merged ===")
        print('')
        for action_log in logs:
            print(action_log)
        if not status:
            print("Resource modifications validation failed")
            sys.exit(1)
        display_warnings(logs, head_commit_msg)

    if args.action == "apply":
        if 'ZUUL_COMMIT' in os.environ:
            print("Read ZUUL_COMMIT from environment: %s" % (
                os.environ['ZUUL_COMMIT']))
            args.zuul_commit = os.environ['ZUUL_COMMIT']
        if not args.zuul_commit:
            print("ZUUL_COMMIT not set. Skip processing.")
            sys.exit(0)
        with TemporaryDirectory() as dpath:
            subprocess.call(
                ["git", "clone", conf.resources.master_repo, dpath])
            os.chdir(dpath)

            # Prev commit may not exists when config repos is not sync push
            if not args.zuul_prev_commit or not is_commit_exists(
                    args.zuul_prev_commit):
                args.zuul_prev_commit = "%s^1" % args.zuul_commit

            print("Checkout at ZUUL COMMIT: %s" % args.zuul_commit)
            subprocess.call(
                ["git", "checkout", args.zuul_commit])
            commit_range = "%s..%s" % (args.zuul_prev_commit, args.zuul_commit)
            head_commit_msg = subprocess.check_output(
                ["git", "log", commit_range]).decode()
            print(head_commit_msg)
            status, logs = engine.apply(
                conf.resources.master_repo, args.zuul_prev_commit,
                dpath, args.zuul_commit)
            print("")
            print(
                "=== Resources actions ===")
            print('')
            for action_log in logs:
                print(action_log)
            if not status:
                print("Resource apply failed")
                sys.exit(1)

    if args.action == "direct-apply":
        if not args.prev_yaml or not args.new_yaml:
            print("Not prev and new yaml. Unable to compute a diff.")
            sys.exit(1)
        status, logs = engine.direct_apply(
            open(args.prev_yaml).read(), open(args.new_yaml).read())
        print("")
        print(
            "=== Resources actions ===")
        print('')
        for action_log in logs:
            print(action_log)
        if not status:
            print("Resource apply failed")
            sys.exit(1)

    if args.action == "remote-validate":
        if not args.remote_gateway:
            print("Need a remote SF gateway URL")
            sys.exit(1)
        if not os.path.isfile('.service_user_password'):
            print("Unable to find .service_user_password")
            sys.exit(1)
        if not args.zuul_prev_commit:
            args.zuul_prev_commit = "master"
        structured_data, head_commit_msg = read_repo_to_validate(
            args.zuul_prev_commit, args.zuul_commit)
        print("=== Remote validation on %s ===" % args.remote_gateway)
        services = get_managesf_info(args.remote_gateway)
        AuthEngine = get_auth_engine(services)
        req_args = AuthEngine.get_request_args(
            args.remote_gateway, username='SF_SERVICE_USER',
            password=open('.service_user_password').read().strip(),
        )
        ret = requests.post(
            args.remote_gateway + '/manage/v2/resources/',
            json={'prev_ref': args.zuul_prev_commit,
                  'data': structured_data},
            **req_args)
        logs = ret.json()
        print("")
        print(
            "=== Resources actions list that will apply once patch merged ===")
        print('')
        for action_log in logs:
            print(action_log)
        if ret.status_code != 200:
            print("Resource modifications validation failed")
            sys.exit(1)
        display_warnings(logs, head_commit_msg)



#!/usr/bin/env python
#
# Copyright (C) 2015  Red Hat <licensing@enovance.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import TestCase
import time

from managesf.services import exceptions as exc
from managesf.services import base
from managesf.tests import dummy_conf


class DummyHooksManager(base.BaseHooksManager):
    def patchset_created(self, *args, **kwargs):
        return 1

    def change_merged(self, *args, **kwargs):
        return 2


class BaseServiceForTest(base.BaseServicePlugin):
    def __init__(self, conf):
        super(BaseServiceForTest, self).__init__(conf)
        self.hooks = DummyHooksManager(self)

    def configure_plugin(self, conf):
        self.conf = conf

    def _get_client(self, cookie=None, **kwargs):
        if time.time() > 1000:
            return "client2"
        else:
            return "client1"


class TestService(TestCase):
    @classmethod
    def setupClass(cls):
        cls.conf = dummy_conf()
        cls.service = BaseServiceForTest(cls.conf)

    @classmethod
    def tearDownClass(cls):
        pass

    def test_is_admin(self):
        self.assertEqual(True,
                         self.service.role.is_admin(self.conf.admin['name']))
        self.assertEqual(False,
                         self.service.role.is_admin('YOLO LMAO'))

    def test_hooks(self):
        self.assertEqual(1,
                         self.service.hooks.patchset_created())
        self.assertEqual(2,
                         self.service.hooks.change_merged())
        self.assertRaises(exc.UnavailableActionError,
                          self.service.hooks.random_undefined_hook,
                          'dummy argument')



# Copyright (c) 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import re
try:
    from urllib import quote_plus
except ImportError:
    from urllib.parse import quote_plus
import hashlib
import logging

from git.config import GitConfigParser

from managesf.services.gerrit import SoftwareFactoryGerrit
from managesf.model.yamlbkd.resource import BaseResource
from managesf.services.gerrit import utils

# ## DEBUG statements to ease run that standalone ###
# import logging
# logging.basicConfig()
# logging.getLogger().setLevel(logging.DEBUG)
# requests_log = logging.getLogger("requests.packages.urllib3")
# requests_log.setLevel(logging.DEBUG)
# requests_log.propagate = True
#
# from pecan import configuration
# from managesf.model.yamlbkd.resources.gitrepository import GitRepositoryOps
# conf = configuration.conf_from_file('/etc/managesf/config.py')
# g = GitRepositoryOps(conf, {})
# g._set_client()
# ###

logger = logging.getLogger(__name__)
DEFAULT_GROUPS = ('Service Users',
                  'Administrators',
                  'Anonymous Users')


class GitRepositoryOps(object):

    def __init__(self, conf, new):
        self.conf = conf
        self.new = new
        self.client = None

    def _set_client(self):
        if not self.client:
            gerrit = SoftwareFactoryGerrit(self.conf)
            self.client = gerrit.get_client()

    def extra_validations(self, **kwargs):
        repo_name = kwargs['name']
        acl_id = kwargs['acl']

        logs = []
        set_private = False

        acls = self.new['resources']['acls'][acl_id]['file']
        # Detect if the ACL make that repository private
        if ('read = deny group Registered Users' in acls and
                'read = deny group Anonymous Users' in acls):
            # Find repos marked as private in projects
            # and remove them from the private_repos accu
            projects = self.new['resources']['projects'].values()
            for project in projects:
                for sr in project['source-repositories']:
                    if isinstance(sr, dict):
                        sr_name = list(sr.keys())[0]
                        sr_attrs = sr[sr_name]
                        if (sr_name == repo_name and
                                sr_attrs.get('private') is True):
                            set_private = True
            if not set_private:
                logs.append(
                    "%s repository use a private Gerrit ACL but are not"
                    " defined as private in a project" % repo_name)

        return logs

    def get_all(self):
        logs = []
        gitrepos = {}
        acls = {}

        self._set_client()

        try:
            repos = self.client.get_projects()
        except Exception as e:
            logger.exception("get_projects failed")
            logs.append("Repo list: err API returned %s" % e)

        for name in repos:
            gitrepos[name] = {}
            r = utils.GerritRepo(name, self.conf)
            # Remove the project section when it only contains description
            remove_project_section = False
            acl_path = r.get_raw_acls()
            acl_groups = set()
            c = GitConfigParser(acl_path)
            c.read()
            for section_name in c.sections():
                for k, v in c.items(section_name):
                    if section_name == 'project':
                        if k == 'description':
                            if len(c.items(section_name)) == 1:
                                remove_project_section = True
                            gitrepos[name]['description'] = v
                        continue
                    r = re.search('group (.*)', v)
                    if r:
                        acl_groups.add(r.groups()[0].strip())

            _acl = open(acl_path).read()
            acl = ""
            # Clean the ACL file to avoid issue at YAML multiline
            # serialization. Remove the description and as a good
            # practice description should never appears in a ACL rtype
            # TODO(fbo): extra_validation of acl must deny the description
            for line in _acl.splitlines():
                if remove_project_section and line.find('[project]') != -1:
                    continue
                if line.find('description') != -1:
                    continue
                acl += line.replace('\t', '    ').rstrip() + '\n'
            m = hashlib.md5()
            m.update(acl.encode())
            acl_id = m.hexdigest()
            gitrepos[name]['name'] = name
            gitrepos[name]['acl'] = acl_id
            acls[acl_id] = {}
            acls[acl_id]['file'] = acl
            acls[acl_id]['groups'] = acl_groups
            acls[acl_id]['groups'] -= set(DEFAULT_GROUPS)
            acls[acl_id]['groups'] -= set(('Registered Users',))
            acls[acl_id]['groups'] = list(acls[acl_id]['groups'])
        return logs, {'repos': gitrepos, 'acls': acls}

    def create(self, **kwargs):
        logs = []
        name = kwargs['name']
        description = kwargs['description']
        kwargs['acl']

        self._set_client()

        try:
            self.client.create_project(name, description, ['Administrators'])
        except Exception as e:
            logger.exception("create_project failed")
            logs.append("Repo create: err API returned %s" % e)

        try:
            r = utils.GerritRepo(name, self.conf)
            r.clone()
        except Exception as e:
            logger.exception("GerritRepo create repo checkout failed")
            logs.append("Repo create fails to checkout the repo %s" % e)
            return logs

        logs.extend(self.install_acl(r, **kwargs))
        logs.extend(self.install_git_review_file(r, name, 'master'))
        logs.extend(self.create_branches(r, **kwargs))

        return logs

    def delete(self, **kwargs):
        logs = []
        name = kwargs['name']

        self._set_client()

        try:
            self.client.delete_project(name, True)
        except Exception as e:
            logger.exception("delete_project failed")
            logs.append("Repo delete: err API returned %s" % e)

        return logs

    def update(self, **kwargs):
        logs = []

        self._set_client()

        try:
            r = utils.GerritRepo(kwargs['name'], self.conf)
            r.clone()
        except Exception as e:
            logger.exception("GerritRepo update repo checkout failed")
            logs.append("Repo update fails to checkout the repo %s" % e)
            return logs

        logs.extend(self.install_acl(r, **kwargs))
        logs.extend(self.create_branches(r, **kwargs))

        return logs

    def install_git_review_file(self, r, name, branch):
        logs = []

        gitreview_template = """[gerrit]
host=%(gerrit-host)s
port=%(gerrit-host-port)s
project=%(name)s
defaultbranch=%(branch)s
"""

        paths = {}
        content = gitreview_template % (
            {'gerrit-host': self.conf.gerrit['top_domain'],
             'gerrit-host-port': self.conf.gerrit['ssh_port'],
             'name': name,
             'branch': branch})
        paths['.gitreview'] = content

        # Clone the master branch and push the .gitreview file
        try:
            r.push_branch(branch, paths)
        except Exception as e:
            logger.exception("GerritRepo push_branch failed")
            logs.append(str(e))

        return logs

    def install_acl(self, r, **kwargs):
        logs = []
        description = kwargs['description']
        acl_id = kwargs['acl']

        group_names = set([])
        acl_data = ""

        # Add default groups implicitly
        for default_group in DEFAULT_GROUPS:
            group_names.add(default_group)

        # If the string is not empty (default empty)
        if acl_id:
            # Fetch the ACL
            acl_data = self.new['resources']['acls'][acl_id]['file']
            acl_group_ids = set(
                self.new['resources']['acls'][acl_id]['groups'])

            # Fetch groups name
            for group_id in acl_group_ids:
                gname = self.new['resources']['groups'][group_id]['name']
                group_names.add(gname)
        else:
            acl_data = """[access]
        inheritFrom = All-Projects
[project]
        description = No description provided"""

        # Fill a groups file
        groups_file = """# UUID Group Name
global:Registered-Users\tRegistered Users"""
        for group in group_names:
            gid = self.client.get_group_id(group)
            groups_file += "\n%s\t%s" % (gid, group)

        # Overwrite the description if given in the ACL file
        if 'description =' in acl_data:
            acl_data = re.sub('description =.*',
                              'description = %s' % description,
                              acl_data)
        else:
            acl_data += """[project]
        description = %s
"""
            acl_data = acl_data % description

        # Clone the meta/config branch and push the ACL
        try:
            paths = {}
            paths['project.config'] = acl_data
            paths['groups'] = groups_file
            r.push_config(paths)
        except Exception as e:
            logger.exception("GerritRepo push_config failed")
            logs.append(str(e))
        return logs

    def set_default_branch(self, name, branch):
        endpoint = "projects/%s/HEAD" % quote_plus(name)
        data = {"ref": "refs/heads/%s" % branch}
        return self.client.put(endpoint, data)

    def create_branches(self, r, **kwargs):
        logs = []
        name = kwargs['name']
        branches = kwargs.get('branches')
        dbranch = kwargs.get('default-branch')

        try:
            refs = r.list_remote_branches()
        except Exception as e:
            logger.exception("GerritRepo create_branches failed")
            logs.append("Get repo branches failed %s" % e)
            return logs

        to_create = dict(
            [(branch, sha) for branch, sha in branches.items() if
             branch not in refs.keys() and sha != '0'])
        to_delete = [branch for branch, sha in branches.items() if
                     branch in refs.keys() and sha == '0']

        if (refs['HEAD'] != dbranch and
                dbranch not in refs.keys() and
                dbranch not in to_create.keys() and
                dbranch != ""):
            # Then a default-branch is requested but not defined in
            # branches list. It may be coherent with the situation of first
            # creating a repo with a specific default-branch
            to_create[dbranch] = 'HEAD'

        # First create requested branches
        for branch, sha in to_create.items():
            try:
                r.create_remote_branch(branch, sha)
                logs.extend(self.install_git_review_file(
                    r, name, branch))
            except Exception as e:
                logger.exception("GerritRepo create_branch/create %s "
                                 "from sha %s failed" % (branch, sha))
                logs.append("Create branch %s from sha %s failed %s" % (
                    branch, sha, e))

        # Set default branch
        if refs['HEAD'] != dbranch and dbranch != "":
            try:
                self.set_default_branch(name, dbranch)
            except Exception as e:
                logger.exception("set_default_branch failed")
                logs.append("Set default branch %s err API returned %s" % (
                            dbranch, e))

        # Then delete requested branches
        for branch in to_delete:
            try:
                r.delete_remote_branch(branch)
            except Exception as e:
                logger.exception("GerritRepo create_branch/delete %s "
                                 "failed" % branch)
                logs.append("Delete branch %s failed %s" % (branch, e))

        return logs


class GitRepository(BaseResource):

    MODEL_TYPE = 'git'
    DESCRIPTION = ("The git resource is used to describe a git repository "
                   "hosted on Gerrit. An acl ID can be provided via the acl "
                   "key.")
    MODEL = {
        'name': (
            str,
            r'^([a-zA-Z0-9\-_\./])+$',
            False,
            "",
            False,
            "The repository name",
        ),
        'description': (
            str,
            '.*',
            False,
            "No description provided",
            True,
            "The repository description",
        ),
        'acl': (
            str,
            '.*',
            False,
            "",
            True,
            "The ACLs id",
        ),
        'default-branch': (
            str,
            r'[a-zA-Z0-9\-_\./]*',
            False,
            "",
            True,
            "The repository default branch. If the branch does not exist yet "
            "or have not been defined in the 'branches' attribute then "
            "default-branch is first created from origin/HEAD",
        ),
        'branches': (
            dict,
            (r'[a-zA-Z0-9\-_\./]+', r'[a-zA-Z0-9\-_\./]+'),
            False,
            {},
            True,
            "Repository branches. Branches name is the key and "
            "branch is the branch value (a SHA/or an existing branch "
            "name/or HEAD). When branch already exist then no reset to "
            "the given value is done except for the '0' value that "
            "ensure the branch does not exist by removing the ref. "
            "If you intend to explicitly declare already existing branches "
            "then please use HEAD as value.",
        ),
    }
    PRIORITY = 20
    PRIMARY_KEY = 'name'
    CALLBACKS = {
        'update': lambda conf, new, kwargs:
            GitRepositoryOps(conf, new).update(**kwargs),
        'create': lambda conf, new, kwargs:
            GitRepositoryOps(conf, new).create(**kwargs),
        'delete': lambda conf, new, kwargs:
            GitRepositoryOps(conf, new).delete(**kwargs),
        'extra_validations': lambda conf, new, kwargs:
            GitRepositoryOps(conf, new).extra_validations(**kwargs),
        'get_all': lambda conf, new:
            GitRepositoryOps(conf, new).get_all(),
    }

    def get_deps(self, keyname=False):
        if keyname:
            return 'acl'
        if self.resource.get('acl'):
            return {'acls': set([self.resource['acl']])}
        else:
            return {'acls': set([])}



#
# Copyright (c) 2018 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

from managesf.model.yamlbkd.resource import BaseResource

logger = logging.getLogger(__name__)


class Tenant(BaseResource):

    DESCRIPTION = ("The tenant resource can be is used to describe a tenant "
                   "like where is located the config repository of tenant.")

    MODEL_TYPE = 'tenant'
    MODEL = {
        'name': (
            str,
            r'^([a-zA-Z0-9\-_\./])+$',
            False,
            "",
            False,
            "The tenant name",
        ),
        'url': (
            str,
            '.*',
            True,
            None,
            True,
            "The Software Factory managesf URL",
        ),
        'tenant-options': (
            dict,
            ('.+', '.+'),
            False,
            {},
            True,
            "List of tenant options",
        ),
        'default-connection': (
            str,
            '.*',
            False,
            "",
            True,
            "The tenant connection to be used in default Zuul pipelines",
        ),
        'allowed-reporters': (
            list,
            '.+',
            False,
            [],
            True,
            "The list of allowed reporter connections for Zuul",
        ),
        'allowed-triggers': (
            list,
            '.+',
            False,
            [],
            True,
            "The list of allowed trigger connections for Zuul",
        ),
        'description': (
            str,
            '.*',
            False,
            "",
            True,
            "The tenant description",
        ),
    }
    PRIORITY = 5
    PRIMARY_KEY = 'name'
    CALLBACKS = {
        'update': lambda conf, new, kwargs: [],
        'create': lambda conf, new, kwargs: [],
        'delete': lambda conf, new, kwargs: [],
        'extra_validations': lambda conf, new, kwargs: [],
        'get_all': lambda conf, new: ([], {}),
    }



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from unittest import TestCase
from mock import patch

from managesf.tests import dummy_conf

from managesf.model.yamlbkd import engine


class EngineRealResourcesTest(TestCase):

    @classmethod
    def setupClass(cls):
        cls.conf = dummy_conf()
        engine.conf = cls.conf

    def test_unknown_resources(self):
        master = {
            'resources': {}
        }
        new = {
            'resources': {
                'abc': {
                }
            }
        }

        with patch('managesf.model.yamlbkd.engine.'
                   'SFResourceBackendEngine._load_resources_data') as lrd, \
                patch('os.path.isdir'), \
                patch('os.mkdir'):
            lrd.return_value = (master, new)
            eng = engine.SFResourceBackendEngine('fake', 'resources')
            valid, logs = eng.validate(None, None, None, None)
            self.assertFalse(valid)
            self.assertEqual(len(logs), 1)
            self.assertIn('Resources type: abc not exists', logs)

    def test_group_validation(self):
        master = {
            'resources': {
                'groups': {
                    'sf/g1': {
                        'description': 'This is a group',
                        'members': [
                            'user1@sftests.com',
                            'user2@sftests.com',
                            ]
                        }
                    }
                }
            }
        new = {
            'resources': {
                'groups': {
                    'sf/g1': {
                        'description': 'This is a cool group',
                        'members': [
                            'user1@sftests.com',
                            'user2@sftests.com',
                            ]
                        }
                    }
                }
            }

        with patch('managesf.model.yamlbkd.engine.'
                   'SFResourceBackendEngine._load_resources_data') as lrd, \
                patch('os.path.isdir'), \
                patch('os.mkdir'), \
                patch('managesf.model.yamlbkd.resources.group.'
                      'GroupOps.extra_validations') as xv:
            lrd.return_value = (master, new)
            xv.return_value = []
            eng = engine.SFResourceBackendEngine('fake', 'resources')
            valid, logs = eng.validate(None, None, None, None)
            self.assertTrue(valid)
            self.assertIn(
                'Resource [type: groups, ID: sf/g1] is going to be updated.',
                logs)
            self.assertEqual(len(logs), 1)

        master = {
            'resources': {
                'groups': {}
                }
            }
        new = {
            'resources': {
                'groups': {
                    'sf/g1': {
                        'description': 'This is a cool group',
                        'members': [
                            'user1@sftests.com',
                            'user2@sftests.com',
                            ]
                        }
                    }
                }
            }

        with patch('managesf.model.yamlbkd.engine.'
                   'SFResourceBackendEngine._load_resources_data') as lrd, \
                patch('os.path.isdir'), \
                patch('os.mkdir'), \
                patch('managesf.model.yamlbkd.resources.group.'
                      'GroupOps.extra_validations') as xv:
            lrd.return_value = (master, new)
            xv.return_value = []
            eng = engine.SFResourceBackendEngine('fake', 'resources')
            valid, logs = eng.validate(None, None, None, None)
            self.assertTrue(valid)
            self.assertIn(
                'Resource [type: groups, ID: sf/g1] is going to be created.',
                logs)
            self.assertEqual(len(logs), 1)

        master = {
            'resources': {
                'groups': {
                    'sf/g1': {
                        'description': 'This is a group',
                        'members': [
                            'user1@sftests.com',
                            'user2@sftests.com',
                            ]
                        }
                    }
                }
            }
        new = {
            'resources': {
                'groups': {
                    'sf/g1': {
                        'description': 'This is a group',
                        'members': [
                            'user2@sftests.com',
                            ]
                        }
                    }
                }
            }
        with patch('managesf.model.yamlbkd.engine.'
                   'SFResourceBackendEngine._load_resources_data') as lrd, \
                patch('os.path.isdir'), \
                patch('os.mkdir'), \
                patch('managesf.model.yamlbkd.resources.group.'
                      'GroupOps.extra_validations') as xv:
            lrd.return_value = (master, new)
            xv.return_value = []
            eng = engine.SFResourceBackendEngine('fake', 'resources')
            valid, logs = eng.validate(None, None, None, None)
            self.assertTrue(valid)
            self.assertIn(
                'Resource [type: groups, ID: sf/g1] is going to be updated.',
                logs)
            self.assertEqual(len(logs), 1)

        master = {
            'resources': {
                'groups': {
                    'sf/g1': {
                        'description': 'This is a group',
                        'members': [
                            'user1@sftests.com',
                            'user2@sftests.com',
                            ]
                        }
                    }
                }
            }
        new = {
            'resources': {
                'groups': {
                    'sf/g2': {
                        'description': 'This is a group',
                        'members': [
                            'user4@sftests.com',
                            ]
                        }
                    }
                }
            }
        with patch('managesf.model.yamlbkd.engine.'
                   'SFResourceBackendEngine._load_resources_data') as lrd, \
                patch('os.path.isdir'), \
                patch('os.mkdir'), \
                patch('managesf.model.yamlbkd.resources.group.'
                      'GroupOps.extra_validations') as xv:
            lrd.return_value = (master, new)
            xv.return_value = []
            eng = engine.SFResourceBackendEngine('fake', 'resources')
            valid, logs = eng.validate(None, None, None, None)
            self.assertTrue(valid)
            self.assertIn(
                'Resource [type: groups, ID: sf/g1] is going to be deleted.',
                logs)
            self.assertIn(
                'Resource [type: groups, ID: sf/g2] is going to be created.',
                logs)
            self.assertEqual(len(logs), 2)

        master = {
            'resources': {
                'groups': {},
                }
            }
        new = {
            'resources': {
                'groups': {
                    'sf/g1': {
                        'description': 'This is a cool group',
                        'members': [
                            'notfound@sftests.com'
                            ]
                        }
                    }
                }
            }
        with patch('managesf.model.yamlbkd.engine.'
                   'SFResourceBackendEngine._load_resources_data') as lrd, \
                patch('os.path.isdir'), \
                patch('os.mkdir'), \
                patch('managesf.model.yamlbkd.resources.group.'
                      'GroupOps.extra_validations') as xv:
            lrd.return_value = (master, new)
            xv.return_value = ['Check group members [notfound@sftests.com '
                               'does not exists]: err API unable to find '
                               'the member']
            eng = engine.SFResourceBackendEngine('fake', 'resources')
            valid, logs = eng.validate(None, None, None, None)
            # The member is not known
            self.assertFalse(valid)
            self.assertIn(
                "Check group members [notfound@sftests.com does not exists]: "
                "err API unable to find the member",
                logs)
            self.assertIn(
                "Resource [type: groups, ID: sf/g1] extra validations failed",
                logs)
            self.assertEqual(len(logs), 2)

        master = {
            'resources': {
                'groups': {}
                }
            }
        new = {
            'resources': {
                'groups': {
                    'Administrators': {
                        'description': 'This is the Admin group',
                        'members': [
                            'user1@sftests.com',
                            'user2@sftests.com',
                            ]
                        }
                    }
                }
            }

        with patch('managesf.model.yamlbkd.engine.'
                   'SFResourceBackendEngine._load_resources_data') as lrd, \
                patch('os.path.isdir'), \
                patch('os.mkdir'), \
                patch('managesf.model.yamlbkd.resources.group.GroupOps.'
                      'check_account_members') as cam:
            lrd.return_value = (master, new)
            cam.return_value = []
            eng = engine.SFResourceBackendEngine('fake', 'resources')
            valid, logs = eng.validate(None, None, None, None)
            self.assertFalse(valid)
            self.assertIn(
                'Check group name [Administrators in not managed '
                'by this API]', logs)
            self.assertIn(
                'Resource [type: groups, ID: Administrators] extra '
                'validations failed', logs)
            self.assertEqual(len(logs), 2)

    def test_deps_inheritage(self):
        # Verifiy a group update does not trigger a full chain update
        master = {
            'resources': {
                'groups': {
                    'sf/g1': {
                        'description': 'This is a group',
                        'members': [
                            'user1@sftests.com',
                            ]
                        }
                    },
                'acls': {
                    'a1': {
                        'file': "this is a\nfake acls",
                        'groups': ['sf/g1'],
                        }
                    },
                'repos': {
                    'r1': {
                        'name': 'sf/r1',
                        'description': 'This is a GIT repo',
                        'acl': 'a1'
                        }
                    },
                'projects': {
                    'p1': {
                        'name': 'p1',
                        'description': 'An awesome project',
                        'source-repositories': ['r1'],
                    },
                }
            }
        }
        new = {
            'resources': {
                'groups': {
                    'sf/g1': {
                        'description': 'This is a group',
                        'members': [
                            'user1@sftests.com',
                            'user2@sftests.com',
                            ]
                        }
                    },
                'acls': {
                    'a1': {
                        'file': "this is a\nfake acls",
                        'groups': ['sf/g1'],
                        }
                    },
                'repos': {
                    'r1': {
                        'name': 'sf/r1',
                        'description': 'This is a GIT repo',
                        'acl': 'a1'
                        }
                    },
                'projects': {
                    'p1': {
                        'name': 'p1',
                        'description': 'An awesome project',
                        'source-repositories': ['r1'],
                    },
                }
            }
        }

        # Verify the ACL update trigger the dependency chain update
        with patch('managesf.model.yamlbkd.engine.'
                   'SFResourceBackendEngine._load_resources_data') as lrd, \
                patch('os.path.isdir'), \
                patch('os.mkdir'), \
                patch('managesf.model.yamlbkd.resources.gitacls.'
                      'ACLOps.extra_validations') as xv, \
                patch('managesf.model.yamlbkd.resources.group.'
                      'GroupOps.extra_validations') as xv2:
            lrd.return_value = (master, new)
            xv.return_value = []
            xv2.return_value = []
            eng = engine.SFResourceBackendEngine('fake', 'resources')
            valid, logs = eng.validate(None, None, None, None)
            self.assertTrue(valid)
            self.assertIn(
                'Resource [type: groups, ID: sf/g1] is going to be updated.',
                logs)
            self.assertEqual(len(logs), 1)

        new = {
            'resources': {
                'groups': {
                    'sf/g1': {
                        'description': 'This is a group',
                        'members': [
                            'user1@sftests.com',
                            ]
                        }
                    },
                'acls': {
                    'a1': {
                        'file': "this is a modified\nfake acls",
                        'groups': ['sf/g1'],
                        }
                    },
                'repos': {
                    'r1': {
                        'name': 'sf/r1',
                        'description': 'This is a GIT repo',
                        'acl': 'a1'
                        }
                    },
                'projects': {
                    'p1': {
                        'name': 'p1',
                        'description': 'An awesome project',
                        'source-repositories': ['r1'],
                    },
                }
            }
        }
        with patch('managesf.model.yamlbkd.engine.'
                   'SFResourceBackendEngine._load_resources_data') as lrd, \
                patch('os.path.isdir'), \
                patch('os.mkdir'), \
                patch('managesf.model.yamlbkd.resources.gitacls.'
                      'ACLOps.extra_validations') as xv, \
                patch('managesf.model.yamlbkd.resources.group.'
                      'GroupOps.extra_validations') as xv2:
            lrd.return_value = (master, new)
            xv.return_value = []
            xv2.return_value = []
            eng = engine.SFResourceBackendEngine('fake', 'resources')
            valid, logs = eng.validate(None, None, None, None)
            self.assertTrue(valid)
            self.assertIn(
                'Resource [type: acls, ID: a1] is going to be updated.',
                logs)
            self.assertIn(
                'Resource [type: repos, ID: r1] need a refresh as at least '
                'one of its dependencies has been updated',
                logs)
            self.assertIn(
                'Resource [type: projects, ID: p1] need a refresh as at least '
                'one of its dependencies has been updated',
                logs)
            self.assertEqual(len(logs), 3)

    def test_acls_validation(self):
        master = {
            'resources': {
                'acls': {}
                }
            }
        new = {
            'resources': {
                'acls': {
                    'a1': {
                        'file': "this is a\nfake acls",
                        'groups': ['g1'],
                        }
                    }
                }
            }
        with patch('managesf.model.yamlbkd.engine.'
                   'SFResourceBackendEngine._load_resources_data') as lrd, \
                patch('os.path.isdir'), \
                patch('os.mkdir'), \
                patch('managesf.model.yamlbkd.resources.gitacls.'
                      'ACLOps.extra_validations') as xv:
            lrd.return_value = (master, new)
            xv.return_value = []
            eng = engine.SFResourceBackendEngine('fake', 'resources')
            valid, logs = eng.validate(None, None, None, None)
            # The group on which the ACLs depends on is missing
            self.assertFalse(valid)
            self.assertIn(
                'Resource [type: acls, ID: a1] depends on an unknown '
                'resource [type: groups, ID: g1]',
                logs)
            self.assertEqual(len(logs), 1)

        master = {
            'resources': {
                'acls': {},
                'groups': {
                    'sf/g1': {
                        'description': 'This is a group',
                        'members': [
                            'user1@sftests.com',
                            'user2@sftests.com',
                            ]
                        }
                    }
                }
            }
        new = {
            'resources': {
                'acls': {
                    'a1': {
                        'file': "this is a\nfake acls",
                        'groups': ['sf/g1'],
                        }
                    },
                'groups': {
                    'sf/g1': {
                        'description': 'This is a group',
                        'members': [
                            'user1@sftests.com',
                            'user2@sftests.com',
                            ]
                        }
                    }
                }
            }
        with patch('managesf.model.yamlbkd.engine.'
                   'SFResourceBackendEngine._load_resources_data') as lrd, \
                patch('os.path.isdir'), \
                patch('os.mkdir'), \
                patch('managesf.model.yamlbkd.resources.gitacls.'
                      'ACLOps.extra_validations') as xv:
            lrd.return_value = (master, new)
            xv.return_value = []
            eng = engine.SFResourceBackendEngine('fake', 'resources')
            valid, logs = eng.validate(None, None, None, None)
            self.assertTrue(valid)
            self.assertIn(
                'Resource [type: acls, ID: a1] is going to be created.',
                logs)
            self.assertEqual(len(logs), 1)

        master = {
            'resources': {
                'acls': {
                    'a1': {
                        'file': "this is a\nfake acls",
                        'groups': ['g1'],
                        }
                    },
                'groups': {
                    'sf/g1': {
                        'description': 'This is a group',
                        'members': [
                            'user1@sftests.com',
                            'user2@sftests.com',
                            ]
                        }
                    }
                }
            }
        new = {
            'resources': {
                'acls': {},
                'groups': {
                    'sf/g1': {
                        'description': 'This is a group',
                        'members': [
                            'user1@sftests.com',
                            'user2@sftests.com',
                            ]
                        }
                    }
                }
            }
        with patch('managesf.model.yamlbkd.engine.'
                   'SFResourceBackendEngine._load_resources_data') as lrd, \
                patch('os.path.isdir'), \
                patch('os.mkdir'), \
                patch('managesf.model.yamlbkd.resources.gitacls.'
                      'ACLOps.extra_validations') as xv:
            lrd.return_value = (master, new)
            xv.return_value = []
            eng = engine.SFResourceBackendEngine('fake', 'resources')
            valid, logs = eng.validate(None, None, None, None)
            self.assertTrue(valid)
            self.assertIn(
                'Resource [type: acls, ID: a1] is going to be deleted.',
                logs)
            self.assertEqual(len(logs), 1)

        master = {
            'resources': {
                'acls': {
                    'a1': {
                        'file': "this is a\nfake acls",
                        'groups': ['sf/g1'],
                        }
                    },
                'groups': {
                    'sf/g1': {
                        'description': 'This is a group',
                        'members': [
                            'user1@sftests.com',
                            'user2@sftests.com',
                            ]
                        }
                    }
                }
            }
        new = {
            'resources': {
                'acls': {
                    'a1': {
                        'file': "this is a\nfake acls",
                        'groups': ['sf/g1'],
                        }
                    },
                'groups': {},
                }
            }
        with patch('managesf.model.yamlbkd.engine.'
                   'SFResourceBackendEngine._load_resources_data') as lrd, \
                patch('os.path.isdir'), \
                patch('os.mkdir'), \
                patch('managesf.model.yamlbkd.resources.group.'
                      'GroupOps.extra_validations') as xv:
            lrd.return_value = (master, new)
            xv.return_value = []
            eng = engine.SFResourceBackendEngine('fake', 'resources')
            valid, logs = eng.validate(None, None, None, None)
            # The group on which the ACLs depends on is missing because
            # it has been removed between master and new
            self.assertFalse(valid)
            self.assertIn(
                'Resource [type: acls, ID: a1] depends on an unknown '
                'resource [type: groups, ID: sf/g1]',
                logs)
            self.assertEqual(len(logs), 1)

        master = {
            'resources': {
                'acls': {},
                'groups': {
                    'sf/g1': {
                        'description': 'This is a group',
                        'members': [
                            'user1@sftests.com',
                            'user2@sftests.com',
                            ]
                        },
                    }
                }
            }
        new = {
            'resources': {
                'acls': {
                    'a1': {
                        'file': """[project]
    description = A description
[access "refs/*"]
    read = group sf/g1
    owner = group sf/g1
[access "refs/for/*"]
    addPatchSet = group sf/g1
[access "refs/heads/*"]
    label-Code-Review = -2..+2 group sf/g1
    label-Verified = -2..+2 group sf/g1
    label-Workflow = -1..+1 group sf/g1
    submit = group sf/g1
    read = group sf/g1
""",
                        'groups': ['sf/g1'],
                        }
                    },
                'groups': {
                    'sf/g1': {
                        'description': 'This is a group',
                        'members': [
                            'user1@sftests.com',
                            'user2@sftests.com',
                            ]
                        },
                    }
                }
            }

        with patch('managesf.model.yamlbkd.engine.'
                   'SFResourceBackendEngine._load_resources_data') as lrd, \
                patch('os.path.isdir'), \
                patch('os.mkdir'), \
                patch('managesf.model.yamlbkd.resources.group.'
                      'GroupOps.extra_validations') as xv:
            lrd.return_value = (master, new)
            xv.return_value = []
            eng = engine.SFResourceBackendEngine('fake', 'resources')
            valid, logs = eng.validate(None, None, None, None)
            self.assertTrue(valid)
            self.assertIn(
                'Resource [type: acls, ID: a1] is going to be created.',
                logs)
            self.assertEqual(len(logs), 1)

        master = {
            'resources': {
                'acls': {},
                'groups': {
                    'sf/g1': {
                        'description': 'This is a group',
                        'members': [
                            'user1@sftests.com',
                            'user2@sftests.com',
                            ]
                        },
                    }
                }
            }
        new = {
            'resources': {
                'acls': {
                    'a1': {
                        'file': """This ACL is
wrong ! This string won't be accepted by Gerrit !
""",
                        'groups': ['sf/g1'],
                        }
                    },
                'groups': {
                    'sf/g1': {
                        'description': 'This is a group',
                        'members': [
                            'user1@sftests.com',
                            'user2@sftests.com',
                            ]
                        },
                    }
                }
            }

        with patch('managesf.model.yamlbkd.engine.'
                   'SFResourceBackendEngine._load_resources_data') as lrd, \
                patch('os.path.isdir'), \
                patch('os.mkdir'), \
                patch('managesf.model.yamlbkd.resources.group.'
                      'GroupOps.extra_validations') as xv:
            lrd.return_value = (master, new)
            xv.return_value = []
            eng = engine.SFResourceBackendEngine('fake', 'resources')
            valid, logs = eng.validate(None, None, None, None)
            # The ACLs is not a valid Git Style config file
            self.assertFalse(valid)
            self.assertTrue(logs[0].startswith(
                "File contains no section headers."))
            self.assertIn(
                'Resource [type: acls, ID: a1] extra validations failed',
                logs)
            self.assertEqual(len(logs), 2)

        master = {
            'resources': {
                'acls': {},
                'groups': {
                    'sf/g1': {
                        'description': 'This is a group',
                        'members': [
                            'user1@sftests.com',
                            'user2@sftests.com',
                            ]
                        },
                    }
                }
            }
        new = {
            'resources': {
                'acls': {
                    'a1': {
                        'file': """[project]
    description = A description
[access "refs/*"]
    read = group sf/g1
    owner = group sf/g1
[access "refs/heads/*"]
    label-Code-Review = -2..+2 group sf/g1
    label-Verified = -2..+2 group sf/g1
    label-Workflow = -1..+1 group sf/g1
    submit = group sf/g2
    read = group sf/g1
""",
                        'groups': ['sf/g1'],
                        }
                    },
                'groups': {
                    'sf/g1': {
                        'description': 'This is a group',
                        'members': [
                            'user1@sftests.com',
                            'user2@sftests.com',
                            ]
                        },
                    }
                }
            }
        with patch('managesf.model.yamlbkd.engine.'
                   'SFResourceBackendEngine._load_resources_data') as lrd, \
                patch('os.path.isdir'), \
                patch('os.mkdir'), \
                patch('managesf.model.yamlbkd.resources.group.'
                      'GroupOps.extra_validations') as xv:
            lrd.return_value = (master, new)
            xv.return_value = []
            eng = engine.SFResourceBackendEngine('fake', 'resources')
            valid, logs = eng.validate(None, None, None, None)
            # sf/g2 is not a known group
            self.assertFalse(valid)
            self.assertIn('ACLs file section (access "refs/heads/*"), key '
                          '(submit) relies on an unknown group name: sf/g2',
                          logs)
            self.assertIn('Resource [type: acls, ID: a1] extra validations '
                          'failed', logs)
            self.assertEqual(len(logs), 2)

        new = {
            'resources': {
                'acls': {
                    'a1': {
                        'file': """[project]
    description = A description
[access "refs/*"]
    read = group sf/g1
    owner = group sf/g1
[access "refs/heads/*"]
    label-Code-Review = -2..+2 group sf/g1
    label-Verified = -2..+2 group sf/g1
    label-Workflow = -1..+1 group sf/g1
    submit = group sf/g2
    read = group sf/g1
""",
                        'groups': ['sf/g1', 'others/g2'],
                        }
                    },
                'groups': {
                    'sf/g1': {
                        'description': 'This is a group',
                        'members': [
                            'user1@sftests.com',
                            'user2@sftests.com',
                            ]
                        },
                    'others/g2': {
                        'description': 'This is a group',
                        'members': [
                            'user1@sftests.com',
                            'user2@sftests.com',
                            ]
                        },
                    }
                }
            }
        with patch('managesf.model.yamlbkd.engine.'
                   'SFResourceBackendEngine._load_resources_data') as lrd, \
                patch('os.path.isdir'), \
                patch('os.mkdir'), \
                patch('managesf.model.yamlbkd.resources.group.'
                      'GroupOps.extra_validations') as xv:
            lrd.return_value = (master, new)
            xv.return_value = []
            eng = engine.SFResourceBackendEngine('fake', 'resources')
            valid, logs = eng.validate(None, None, None, None)
            # others/g2 is not a known group
            self.assertFalse(valid)
            self.assertIn('ACLs file section (access "refs/heads/*"), key '
                          '(submit) relies on an unknown group name: sf/g2',
                          logs)
            self.assertIn('Resource [type: acls, ID: a1] extra validations '
                          'failed', logs)

        new = {
            'resources': {
                'acls': {
                    'a1': {
                        'file': """[project]
    description = A description
[access "refs/*"]
    read = group sf/g1
    owner = group sf/g1
[access "refs/heads/*"]
    label-Code-Review = -2..+2 group sf/g1
    label-Verified = -2..+2 group sf/g1
    label-Workflow = -1..+1 group sf/g1
    submit = group g2
    read = group sf/g1
""",
                        'groups': ['sf/g1', 'g2'],
                        }
                    },
                'groups': {
                    'sf/g1': {
                        'description': 'This is a group',
                        'members': [
                            'user1@sftests.com',
                            'user2@sftests.com',
                            ]
                        },
                    'g2': {
                        'description': 'This is a group',
                        'members': [
                            'user1@sftests.com',
                            'user2@sftests.com',
                            ]
                        },
                    }
                }
            }
        with patch('managesf.model.yamlbkd.engine.'
                   'SFResourceBackendEngine._load_resources_data') as lrd, \
                patch('os.path.isdir'), \
                patch('os.mkdir'), \
                patch('managesf.model.yamlbkd.resources.group.'
                      'GroupOps.extra_validations') as xv:
            lrd.return_value = (master, new)
            xv.return_value = []
            eng = engine.SFResourceBackendEngine('fake', 'resources')
            valid, logs = eng.validate(None, None, None, None)
            self.assertTrue(valid)
            self.assertIn(
                'Resource [type: acls, ID: a1] is going to be created.',
                logs)

        master = {
            'resources': {
                'acls': {
                    'a1': {
                        'file': """[project]
    description = A description
[access "refs/*"]
    read = group Registered Users
    read = group Anonymous Users
""",
                        'groups': [],
                        }
                    },
                'groups': {},
                'projects': {},
                'repos': {
                    'r1': {
                        'acl': 'a1',
                    }
                }
            }
        }

        new = {
            'resources': {
                'acls': {
                    'a1': {
                        'file': """[project]
    description = A description
[access "refs/*"]
    read = deny group Registered Users
    read = deny group Anonymous Users
""",
                        'groups': [],
                        }
                    },
                'groups': {},
                'projects': {},
                'repos': {
                    'r1': {
                        'acl': 'a1',
                        }
                    }
                }
            }
        with patch('managesf.model.yamlbkd.engine.'
                   'SFResourceBackendEngine._load_resources_data') as lrd, \
                patch('os.path.isdir'), \
                patch('os.mkdir'), \
                patch('managesf.model.yamlbkd.resources.group.'
                      'GroupOps.extra_validations') as xv:
            lrd.return_value = (master, new)
            xv.return_value = []
            eng = engine.SFResourceBackendEngine('fake', 'resources')
            valid, logs = eng.validate(None, None, None, None)
            self.assertFalse(valid)
            self.assertIn(
                "['r1'] repositories use a private Gerrit ACL but are not "
                "defined as private in a project", logs)
            self.assertIn(
                'Resource [type: acls, ID: a1] extra validations failed', logs)

        new = {
            'resources': {
                'acls': {
                    'a1': {
                        'file': """[project]
    description = A description
    [access "refs/*"]
    read = deny group Registered Users
    read = deny group Anonymous Users
    """,
                        'groups': [],
                        }
                    },
                'groups': {},
                'projects': {
                    'p1': {
                        'description': 'desc',
                        'source-repositories': [
                            {'r1': {'private': True}},
                        ]
                    }
                },
                'repos': {
                    'r1': {
                        'acl': 'a1',
                        }
                    }
                }
            }
        with patch('managesf.model.yamlbkd.engine.'
                   'SFResourceBackendEngine._load_resources_data') as lrd, \
                patch('os.path.isdir'), \
                patch('os.mkdir'), \
                patch('managesf.model.yamlbkd.resources.group.'
                      'GroupOps.extra_validations') as xv:
            lrd.return_value = (master, new)
            xv.return_value = []
            eng = engine.SFResourceBackendEngine('fake', 'resources')
            valid, logs = eng.validate(None, None, None, None)
            self.assertTrue(valid)
            self.assertEqual(len(logs), 4)

        master = {
            'resources': {
                'acls': {
                    'public': {
                        'file': """[project]
    description = A description
[access "refs/*"]
    read = group Registered Users
    read = group Anonymous Users
""",
                        'groups': [],
                        },
                    'private': {
                        'file': """[project]
    description = A description
[access "refs/*"]
    read = deny group Registered Users
    read = deny group Anonymous Users
""",
                        'groups': [],
                        }
                    },
                'groups': {},
                'projects': {},
                'repos': {
                    'r1': {
                        'acl': 'public',
                    }
                }
            }
        }

        new = {
            'resources': {
                'acls': {
                    'public': {
                        'file': """[project]
    description = A description
[access "refs/*"]
    read = group Registered Users
    read = group Anonymous Users
""",
                        'groups': [],
                        },
                    'private': {
                        'file': """[project]
    description = A description
[access "refs/*"]
    read = deny group Registered Users
    read = deny group Anonymous Users
""",
                        'groups': [],
                        }
                    },
                'groups': {},
                'projects': {},
                'repos': {
                    'r1': {
                        'acl': 'private',
                    }
                }
            }
        }

        with patch('managesf.model.yamlbkd.engine.'
                   'SFResourceBackendEngine._load_resources_data') as lrd, \
                patch('os.path.isdir'), \
                patch('os.mkdir'), \
                patch('managesf.model.yamlbkd.resources.group.'
                      'GroupOps.extra_validations') as xv:
            lrd.return_value = (master, new)
            xv.return_value = []
            eng = engine.SFResourceBackendEngine('fake', 'resources')
            valid, logs = eng.validate(None, None, None, None)
            self.assertFalse(valid)
            self.assertIn(
                "r1 repository use a private Gerrit ACL but are not "
                "defined as private in a project", logs)
            self.assertIn(
                'Resource [type: repos, ID: r1] extra validations failed',
                logs)

        new = {
            'resources': {
                'acls': {
                    'public': {
                        'file': """[project]
    description = A description
[access "refs/*"]
    read = group Registered Users
    read = group Anonymous Users
""",
                        'groups': [],
                        },
                    'private': {
                        'file': """[project]
    description = A description
[access "refs/*"]
    read = deny group Registered Users
    read = deny group Anonymous Users
""",
                        'groups': [],
                        }
                    },
                'groups': {},
                'projects': {
                    'p1': {
                        'description': 'desc',
                        'source-repositories': [
                            {'r1': {'private': True}},
                        ]
                    }
                },
                'repos': {
                    'r1': {
                        'acl': 'private',
                    }
                }
            }
        }

        with patch('managesf.model.yamlbkd.engine.'
                   'SFResourceBackendEngine._load_resources_data') as lrd, \
                patch('os.path.isdir'), \
                patch('os.mkdir'), \
                patch('managesf.model.yamlbkd.resources.group.'
                      'GroupOps.extra_validations') as xv:
            lrd.return_value = (master, new)
            xv.return_value = []
            eng = engine.SFResourceBackendEngine('fake', 'resources')
            valid, logs = eng.validate(None, None, None, None)
            self.assertTrue(valid)

        master = {
            'resources': {
                'acls': {
                    'public': {
                        'file': """[project]
    description = A description
[access "refs/*"]
    read = group Registered Users
    read = group Anonymous Users
""",
                        'groups': [],
                        },
                    'private': {
                        'file': """[project]
    description = A description
[access "refs/*"]
    read = deny group Registered Users
    read = deny group Anonymous Users
""",
                        'groups': [],
                        }
                    },
                'groups': {},
                'projects': {},
                'repos': {
                    'r1': {
                        'acl': 'private',
                    }
                }
            }
        }

        new = {
            'resources': {
                'acls': {
                    'public': {
                        'file': """[project]
    description = A description
[access "refs/*"]
    read = group Registered Users
    read = group Anonymous Users
""",
                        'groups': [],
                        },
                    'private': {
                        'file': """[project]
    description = A description
[access "refs/*"]
    read = deny group Registered Users
    read = deny group Anonymous Users
""",
                        'groups': [],
                        }
                    },
                'groups': {},
                'projects': {
                    'p1': {
                        'description': 'desc',
                        'source-repositories': [
                            {'r1': {}}
                        ]
                    }
                },
                'repos': {
                    'r1': {
                        'acl': 'private',
                    }
                }
            }
        }

        with patch('managesf.model.yamlbkd.engine.'
                   'SFResourceBackendEngine._load_resources_data') as lrd, \
                patch('os.path.isdir'), \
                patch('os.mkdir'), \
                patch('managesf.model.yamlbkd.resources.group.'
                      'GroupOps.extra_validations') as xv:
            lrd.return_value = (master, new)
            xv.return_value = []
            eng = engine.SFResourceBackendEngine('fake', 'resources')
            valid, logs = eng.validate(None, None, None, None)
            self.assertFalse(valid)
            self.assertIn(
                "r1 repository use a private Gerrit ACL but are not "
                "defined as private in the project p1", logs)
            self.assertIn(
                'Resource [type: projects, ID: p1] extra validations failed',
                logs)

        new = {
            'resources': {
                'acls': {
                    'public': {
                        'file': """[project]
    description = A description
[access "refs/*"]
    read = group Registered Users
    read = group Anonymous Users
""",
                        'groups': [],
                        },
                    'private': {
                        'file': """[project]
    description = A description
[access "refs/*"]
    read = deny group Registered Users
    read = deny group Anonymous Users
""",
                        'groups': [],
                        }
                    },
                'groups': {},
                'projects': {
                    'p1': {
                        'description': 'desc',
                        'source-repositories': [
                            {'r1': {'private': True}}
                        ]
                    }
                },
                'repos': {
                    'r1': {
                        'acl': 'private',
                    }
                }
            }
        }

        with patch('managesf.model.yamlbkd.engine.'
                   'SFResourceBackendEngine._load_resources_data') as lrd, \
                patch('os.path.isdir'), \
                patch('os.mkdir'), \
                patch('managesf.model.yamlbkd.resources.group.'
                      'GroupOps.extra_validations') as xv:
            lrd.return_value = (master, new)
            xv.return_value = []
            eng = engine.SFResourceBackendEngine('fake', 'resources')
            valid, logs = eng.validate(None, None, None, None)
            self.assertTrue(valid)

    def test_gitrepo_validation(self):
        master = {
            'resources': {
                'acls': {
                    'a1': {
                        'file': 'fake',
                        }
                    },
                'repos': {},
                }
            }
        new = {
            'resources': {
                'repos': {
                    'sf/r1': {
                        'description': 'This is a GIT repo',
                        'acl': 'a1'
                        }
                    },
                'acls': {
                    'a1': {
                        'file': 'fake',
                        }
                    }
                }
            }
        with patch('managesf.model.yamlbkd.engine.'
                   'SFResourceBackendEngine._load_resources_data') as lrd, \
                patch('os.path.isdir'), \
                patch('os.mkdir'), \
                patch('managesf.model.yamlbkd.resources.gitacls.'
                      'ACLOps.extra_validations') as xv:
            lrd.return_value = (master, new)
            xv.return_value = []
            eng = engine.SFResourceBackendEngine('fake', 'resources')
            valid, logs = eng.validate(None, None, None, None)
            self.assertTrue(valid)
            self.assertIn('Resource [type: repos, ID: sf/r1] is going to '
                          'be created.',
                          logs)
            self.assertEqual(len(logs), 1)

        master = {
            'resources': {
                'repos': {
                    'sf/r1': {
                        'description': 'This is a GIT repo',
                        'acl': 'a1'
                        }
                    },
                'acls': {
                    'a1': {
                        'file': 'fake',
                        }
                    },
                }
            }
        new = {
            'resources': {
                'repos': {
                    'sf/r1': {
                        'description': 'This is a GIT repo',
                        'acl': 'a1'
                        }
                    },
                'acls': {
                    'a1': {
                        'file': 'fake2',
                        }
                    }
                }
            }

        with patch('managesf.model.yamlbkd.engine.'
                   'SFResourceBackendEngine._load_resources_data') as lrd, \
                patch('os.path.isdir'), \
                patch('os.mkdir'), \
                patch('managesf.model.yamlbkd.resources.gitacls.'
                      'ACLOps.extra_validations') as xv:
            lrd.return_value = (master, new)
            xv.return_value = []
            eng = engine.SFResourceBackendEngine('fake', 'resources')
            valid, logs = eng.validate(None, None, None, None)
            self.assertTrue(valid)
            self.assertIn('Resource [type: acls, ID: a1] is going to '
                          'be updated.', logs)
            self.assertIn('Resource [type: repos, ID: sf/r1] need a refresh '
                          'as at least one of its dependencies has been '
                          'updated', logs)
            self.assertEqual(len(logs), 2)

        master = {
            'resources': {
                'repos': {
                    'sf/r1': {
                        'description': 'This is a GIT repo',
                        'acl': 'a1'
                        }
                    },
                'acls': {
                    'a1': {
                        'file': 'fake',
                        }
                    },
                }
            }

        new = {
            'resources': {
                'repos': {
                    'sf/r1': {
                        'description': 'This is a GIT repo',
                        'acl': 'a1'
                        }
                    },
                'acls': {},
                }
            }
        with patch('managesf.model.yamlbkd.engine.'
                   'SFResourceBackendEngine._load_resources_data') as lrd, \
                patch('os.path.isdir'), \
                patch('os.mkdir'), \
                patch('managesf.model.yamlbkd.resources.gitacls.'
                      'ACLOps.extra_validations') as xv:
            lrd.return_value = (master, new)
            xv.return_value = []
            eng = engine.SFResourceBackendEngine('fake', 'resources')
            # GIT repository relie on an unknown ACL
            valid, logs = eng.validate(None, None, None, None)
            self.assertFalse(valid)
            self.assertIn('Resource [type: repos, ID: sf/r1] depends on '
                          'an unknown resource [type: acls, ID: a1]', logs)
            self.assertEqual(len(logs), 1)

        master = {
            'resources': {
                'acls': {
                    'a1': {
                        'file': 'fake',
                        }
                    },
                'repos': {},
                }
            }
        new = {
            'resources': {
                'repos': {
                    'sf/r1': {
                        'description': 'This is a GIT repo',
                        'acl': 'a1',
                        'branches': {
                            'br1': 'aa65de9c3eb5496e3119a73f732ebb142753ed54'
                            }
                        }
                    },
                'acls': {
                    'a1': {
                        'file': 'fake',
                        }
                    }
                }
            }
        with patch('managesf.model.yamlbkd.engine.'
                   'SFResourceBackendEngine._load_resources_data') as lrd, \
                patch('os.path.isdir'), \
                patch('os.mkdir'), \
                patch('managesf.model.yamlbkd.resources.gitacls.'
                      'ACLOps.extra_validations') as xv:
            lrd.return_value = (master, new)
            xv.return_value = []
            eng = engine.SFResourceBackendEngine('fake', 'resources')
            valid, logs = eng.validate(None, None, None, None)
            self.assertTrue(valid)
            self.assertIn('Resource [type: repos, ID: sf/r1] is going to '
                          'be created.',
                          logs)
            self.assertEqual(len(logs), 1)

    def test_project_validation(self):
        master = {
            'resources': {
                'projects': {},
                'groups': {},
                'acls': {
                    'a1': {
                        'file': 'fake',
                        }
                    },
                'repos': {
                    'r1': {
                        'name': 'sf/r1',
                        'description': 'This is a GIT repo',
                        'acl': 'a1'
                        }
                    }
                }
            }
        new = {
            'resources': {
                'projects': {
                    'p1': {
                        'name': 'p1',
                        'description': 'An awesome project',
                        'tenant': 'local',
                        'connection': 'github.com',
                        'source-repositories': ['r1'],
                    },
                },
                'groups': {},
                'repos': {
                    'r1': {
                        'name': 'sf/r1',
                        'description': 'This is a GIT repo',
                        'acl': 'a1'
                        }
                    },
                'acls': {
                    'a1': {
                        'file': 'fake',
                        }
                    }
                }
            }

        with patch('managesf.model.yamlbkd.engine.'
                   'SFResourceBackendEngine._load_resources_data') as lrd, \
                patch('os.path.isdir'), \
                patch('os.mkdir'), \
                patch('managesf.model.yamlbkd.resources.gitacls.'
                      'ACLOps.extra_validations') as xv, \
                patch('managesf.model.yamlbkd.resources.group.'
                      'GroupOps.extra_validations') as xv2:
            lrd.return_value = (master, new)
            xv.return_value = []
            eng = engine.SFResourceBackendEngine('fake', 'resources')
            valid, logs = eng.validate(None, None, None, None)
            self.assertTrue(valid)
            self.assertIn('Resource [type: projects, ID: p1] is going to '
                          'be created.', logs)
            self.assertEqual(len(logs), 1)

        new = {
            'resources': {
                'projects': {
                    'p1': {
                        'name': 'p1',
                        'description': 'An awesome project',
                        'source-repositories': ['r1', 'r2'],
                    },
                },
                'groups': {},
                'repos': {
                    'r1': {
                        'name': 'sf/r1',
                        'description': 'This is a GIT repo',
                        'acl': 'a1'
                        }
                    },
                'acls': {
                    'a1': {
                        'file': 'fake',
                        }
                    }
                }
            }
        with patch('managesf.model.yamlbkd.engine.'
                   'SFResourceBackendEngine._load_resources_data') as lrd, \
                patch('os.path.isdir'), \
                patch('os.mkdir'), \
                patch('managesf.model.yamlbkd.resources.gitacls.'
                      'ACLOps.extra_validations') as xv, \
                patch('managesf.model.yamlbkd.resources.group.'
                      'GroupOps.extra_validations') as xv2:
            lrd.return_value = (master, new)
            xv.return_value = []
            eng = engine.SFResourceBackendEngine('fake', 'resources')
            valid, logs = eng.validate(None, None, None, None)
            # Project depends on an unknown r2 resource but soft deps is
            # set to True for project type so no deps check is performed
            # between projects and repos.
            self.assertTrue(valid)

        master = {
            'resources': {
                'projects': {},
                }
            }

        new = {
            'resources': {
                'projects': {
                    'p1': {
                        'name': 'p1',
                        'description': 'An awesome project',
                        'options': [
                            'repoxplorer/skip',
                        ],
                        'source-repositories': [
                            {'config': {
                                'zuul/config-project': True
                            }},
                            'sf-jobs',
                            'zuul-jobs',
                        ],
                    }
                }
            }
        }

        with patch('managesf.model.yamlbkd.engine.'
                   'SFResourceBackendEngine._load_resources_data') as lrd, \
                patch('os.path.isdir'), \
                patch('os.mkdir'), \
                patch('managesf.model.yamlbkd.resources.gitacls.'
                      'ACLOps.extra_validations') as xv, \
                patch('managesf.model.yamlbkd.resources.group.'
                      'GroupOps.extra_validations') as xv2:
            lrd.return_value = (master, new)
            xv.return_value = []
            eng = engine.SFResourceBackendEngine('fake', 'resources')
            valid, logs = eng.validate(None, None, None, None)
            self.assertTrue(valid)
            self.assertIn('Resource [type: projects, ID: p1] is going to '
                          'be created.', logs)

        master = {
            'resources': {
                'projects': {
                    'p1': {
                        'name': 'p1',
                        'description': 'An awesome project',
                        'source-repositories': [
                            {'config': {
                                'zuul/config-project': True
                            }},
                            'sf-jobs',
                            'zuul-jobs',
                        ],
                    }
                }
            }
        }

        new = {
            'resources': {
                'projects': {
                    'p1': {
                        'name': 'p1',
                        'description': 'An awesome project',
                        'source-repositories': [
                            {'config2': {
                                'zuul/config-project': True
                            }},
                            'sf-jobs',
                            'zuul-jobs',
                        ],
                    }
                }
            }
        }

        with patch('managesf.model.yamlbkd.engine.'
                   'SFResourceBackendEngine._load_resources_data') as lrd, \
                patch('os.path.isdir'), \
                patch('os.mkdir'), \
                patch('managesf.model.yamlbkd.resources.gitacls.'
                      'ACLOps.extra_validations') as xv, \
                patch('managesf.model.yamlbkd.resources.group.'
                      'GroupOps.extra_validations') as xv2:
            lrd.return_value = (master, new)
            xv.return_value = []
            eng = engine.SFResourceBackendEngine('fake', 'resources')
            valid, logs = eng.validate(None, None, None, None)
            self.assertTrue(valid)
            self.assertIn('Resource [type: projects, ID: p1] is going to '
                          'be updated.', logs)

        new = {
            'resources': {
                'projects': {
                    'p1': {
                        'name': 'p1',
                        'description': 'An awesome project',
                        'source-repositories': [
                            {'config': {
                                'zuul/config-project': True,
                                'default-branch': ['devel', 'stable-2.4'],
                            }},
                            'sf-jobs',
                            'zuul-jobs',
                        ],
                    }
                }
            }
        }

        with patch('managesf.model.yamlbkd.engine.'
                   'SFResourceBackendEngine._load_resources_data') as lrd, \
                patch('os.path.isdir'), \
                patch('os.mkdir'), \
                patch('managesf.model.yamlbkd.resources.gitacls.'
                      'ACLOps.extra_validations') as xv, \
                patch('managesf.model.yamlbkd.resources.group.'
                      'GroupOps.extra_validations') as xv2:
            lrd.return_value = (master, new)
            xv.return_value = []
            eng = engine.SFResourceBackendEngine('fake', 'resources')
            valid, logs = eng.validate(None, None, None, None)
            self.assertTrue(valid)
            self.assertIn('Resource [type: projects, ID: p1] is going to '
                          'be updated.', logs)

        master = {
            'resources': {
                'projects': {
                    'p1': {
                        'name': 'p1',
                        'description': 'An awesome project',
                        'source-repositories': ['r1'],
                    },
                },
                'groups': {},
                'acls': {
                    'a1': {
                        'file': 'fake',
                        }
                    },
                'repos': {
                    'r1': {
                        'name': 'sf/r1',
                        'description': 'This is a GIT repo',
                        'acl': 'a1'
                        }
                    }
                }
            }
        new = {
            'resources': {
                'projects': {
                    'p1': {
                        'name': 'p1',
                        'description': 'An awesome project',
                        'source-repositories': ['r1'],
                    },
                },
                'repos': {
                    'r1': {
                        'name': 'sf/r1',
                        'description': 'This is a GIT repo',
                        'acl': 'a1'
                        }
                    },
                'acls': {
                    'a1': {
                        'file': 'fake',
                        'groups': ['g1'],
                        }
                    },
                'groups': {
                    'g1': {
                        'name': 'sf/g1',
                        'members': [],
                    },
                }
            }
        }
        with patch('managesf.model.yamlbkd.engine.'
                   'SFResourceBackendEngine._load_resources_data') as lrd, \
                patch('os.path.isdir'), \
                patch('os.mkdir'), \
                patch('managesf.model.yamlbkd.resources.gitacls.'
                      'ACLOps.extra_validations') as xv, \
                patch('managesf.model.yamlbkd.resources.group.'
                      'GroupOps.extra_validations') as xv2:
            lrd.return_value = (master, new)
            xv.return_value = []
            xv2.return_value = []
            eng = engine.SFResourceBackendEngine('fake', 'resources')
            valid, logs = eng.validate(None, None, None, None)
            self.assertTrue(valid)
            self.assertIn('Resource [type: acls, ID: a1] is going '
                          'to be updated.', logs)
            self.assertIn('Resource [type: projects, ID: p1] need a '
                          'refresh as at least one of its dependencies '
                          'has been updated', logs)
            self.assertIn('Resource [type: groups, ID: g1] is '
                          'going to be created.', logs)
            self.assertIn('Resource [type: repos, ID: r1] need a '
                          'refresh as at least one of its dependencies '
                          'has been updated', logs)
            self.assertEqual(len(logs), 4)

    def test_tenant_validation(self):
        master = {
            'resources': {
                'tenants': {}
            }
        }
        new = {
            'resources': {
                'tenants': {
                    'kimchi': {
                        'url': 'https://kimchi.sftests.com/manage',
                        'tenant-options': {
                            'zuul/default-jobs-timeout': 3600,
                            'zuul/exclude-unprotected-branches': True,
                            'zuul/default-parent': 'myparent',
                            }
                        },
                    'natto': {
                        'url': 'https://natto.sftests.com/manage',
                        'description': 'The natto tenant',
                        'default-connection': 'gerrit',
                        'allowed-triggers': ['gerrit1', 'gerrit2'],
                        'allowed-reporters': ['review.openstack.org'],
                        },
                    }
                }
            }

        with patch('managesf.model.yamlbkd.engine.'
                   'SFResourceBackendEngine._load_resources_data') as lrd, \
                patch('os.path.isdir'), \
                patch('os.mkdir'):
            lrd.return_value = (master, new)
            eng = engine.SFResourceBackendEngine('fake', 'resources')
            valid, logs = eng.validate(None, None, None, None)
            self.assertTrue(valid)
            self.assertIn('Resource [type: tenants, ID: kimchi] is going to '
                          'be created.', logs)
            self.assertIn('Resource [type: tenants, ID: natto] is going to '
                          'be created.', logs)
            self.assertEqual(len(logs), 2)

    def test_connection_validation(self):
        master = {
            'resources': {
                'connections': {
                    'local': {
                        'base-url': 'https://sftests.com/r/',
                        'type': 'gerrit',
                        },
                    }
            }
        }
        new = {
            'resources': {
                'connections': {
                    'local': {
                        'base-url': 'https://sftests.com/r/',
                        'type': 'gerrit',
                        },
                    'github.com': {
                        'github-app-name': 'softwarefactory-project-zuul',
                        'github-label': 'zuulit',
                        'type': 'github'
                        },
                    }
                }
            }

        with patch('managesf.model.yamlbkd.engine.'
                   'SFResourceBackendEngine._load_resources_data') as lrd, \
                patch('os.path.isdir'), \
                patch('os.mkdir'):
            lrd.return_value = (master, new)
            eng = engine.SFResourceBackendEngine('fake', 'resources')
            valid, logs = eng.validate(None, None, None, None)
            self.assertFalse(valid)
            self.assertIn("Connections can't be updated manually, they are "
                          "managed by configuration management.", logs)
            self.assertEqual(len(logs), 2)

    def test_get_missing_resources(self):
        eng = engine.SFResourceBackendEngine(None, None)
        current_resources = {
            'resources': {
                'projects': {
                    'p1': {
                        'description': 'An awesome project',
                        'source-repositories': ['r1'],
                    },
                },
                'groups': {
                    'g1': {
                        'members': ['user2@sftests.com'],
                    },
                    'g2': {
                        'members': ['user3@sftests.com'],
                    },
                },
                'repos': {
                    'sf/r1': {
                        'acl': 'a1',
                    },
                    'sf/r2': {
                        'acl': 'a1',
                    },
                },
                'acls': {
                    'a1': {
                        'file': 'fake',
                        'groups': [],
                    }
                }
            }
        }
        gr_reality = {
            'repos': {
                'sf/r1': {
                    'acl': 'hash77',
                },
                'sf/r2': {
                    'acl': 'hash77',
                },
                'sf/r3': {
                    'acl': 'hash77',
                },
            },
            'acls': {
                'hash77': {
                    'file': 'fake',
                    'groups': [],
                },
                'hash78': {
                    'file': 'fake2',
                    'groups': [],
                },
            },
        }
        g_reality = {
            'groups': {
                'g3': {
                    'members': ['user3@sftests.com'],
                },
            },
        }
        expected = {
            'resources': {
                'groups': {
                    'g3': {
                        'members': ['user3@sftests.com']
                    },
                },
                'repos': {
                    'sf/r3': {
                        'acl': 'a1',
                    }
                },
                'acls': {
                    'hash78': {
                        'groups': [],
                        'file': 'fake2',
                    }
                },
            }
        }
        with patch('managesf.model.yamlbkd.engine.'
                   'SFResourceBackendEngine.get') as g, \
                patch('managesf.model.yamlbkd.resources.gitrepository.'
                      'GitRepositoryOps.get_all') as gar, \
                patch('managesf.model.yamlbkd.resources.group.'
                      'GroupOps.get_all') as gag:
            gar.return_value = ([], gr_reality)
            gag.return_value = ([], g_reality)
            g.return_value = current_resources
            logs, tree = eng.get_missing_resources(None, None)
            self.assertListEqual(logs, [])
            self.assertDictEqual(tree, expected)

    def test_resources_project_get(self):
        master = {
            'resources': {
                'projects': {
                    'p1': {
                        'name': 'p1',
                        'description': 'An awesome project',
                        'source-repositories': [
                            'r1',
                            {'r2': {
                                'zuul/config-project': True,
                                'default-branch': 'main'
                            }}
                        ],
                    },
                }
            }
        }

        with patch('managesf.model.yamlbkd.engine.YAMLBackend.'
                   'get_data') as gd, \
                patch('managesf.model.yamlbkd.engine.YAMLBackend.refresh'), \
                patch('os.path.isdir'), \
                patch('os.mkdir'):
            gd.return_value = master
            eng = engine.SFResourceBackendEngine('fake', 'resources')
            data = eng.get(None, None)
            expected = {
                'config-repo': None,
                'public-url': 'http://sftests.com/manage',
                'resources': {
                    'projects': {
                        'p1': {
                            'source-repositories': [
                                {'r1': {}},
                                {'r2': {'zuul/config-project': True,
                                        'default-branch': 'main'}}
                            ],
                            'description': 'An awesome project',
                            'name': 'p1'}
                        }
                    }
                }
            self.assertDictEqual(data, expected)



#!/usr/bin/env python
#
# Copyright (C) 2015 Red Hat <licensing@enovance.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


class ServiceNotAvailableError(Exception):
    """Raised if a service plugin cannot be configured"""
    pass


class UnavailableActionError(Exception):
    """Raised if a service does not know how to do the requested action"""
    pass


class Unauthorized(Exception):
    """Raised if an action is unauthorized on the service"""
    pass


class CreateGroupException(Exception):
    """Raised if a group creation failed"""
    pass


class UpdateGroupException(Exception):
    """Raised if a group update failed"""
    pass


class GroupNotFoundException(Exception):
    """Raised if a group lookup failed"""
    pass



#
# Copyright (c) 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import git
import logging
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO
import yaml

try:
    from yaml import CSafeDumper as YDumper
    from yaml import CSafeLoader as YLoader
except ImportError:
    from yaml import SafeDumper as YDumper
    from yaml import SafeLoader as YLoader

from pecan import conf  # noqa


logger = logging.getLogger(__name__)

RESOURCES_STRUCT = {'resources': {'rtype': {'key': {}}}}


class YAMLDBException(Exception):
    pass


class YAMLBackend(object):
    def __init__(self, git_repo_url, git_ref, sub_dir,
                 clone_path, cache_path):
        """ Class to read and validate resources from YAML
        files from stored in a GIT repository. The main data
        structure as well as resources structure must follow
        specific constraints.

        This Class also maintains a cache file to avoid full
        re-load and validation at init when ref hash has not
        changed.

        :param git_repo_url: The URI of the GIT repository
        :param git_ref: The GIT repository refs such as
               refs/zuul/master/Z3a46b05b4574472bbf093ff5562cba42
               or refs/heads/master
        :param sub_dir: The path from the GIT root to YAML files
        :param clone_path: The path where to clone the GIT repository
        :param cache_path: The path to the cached file
        """
        self._initialize(git_repo_url, git_ref, sub_dir,
                         clone_path, cache_path)
        self.refresh()

    def _initialize(self, git_repo_url, git_ref, sub_dir,
                    clone_path, cache_path):
        self.git_repo_url = git_repo_url
        self.git_ref = git_ref
        self.clone_path = clone_path
        self.cache_path = cache_path
        self.cache_path_hash = "%s_%s" % (cache_path, '_hash')
        self.db_path = os.path.join(self.clone_path, sub_dir)
        self.rids = {}

    def _get_repo_hash(self):
        repo = git.Git(self.clone_path)
        repo_hash = repo.execute(['git', '--no-pager', 'log', '-1',
                                  '--pretty=%h', 'HEAD'])
        return repo_hash

    def _get_cache_hash(self):
        return open(self.cache_path_hash).read().strip()

    def _update_cache(self):
        repo_hash = self._get_repo_hash()
        self.hash = repo_hash
        yaml.dump(self.data, open(self.cache_path, 'w'),
                  Dumper=YDumper)
        open(self.cache_path_hash, 'w').write(repo_hash)
        logger.info("Cache file has been updated.")

    def _load_from_cache(self):
        if not os.path.isfile(self.cache_path_hash):
            logger.info("No DB cache file found.")
        else:
            repo_hash = self._get_repo_hash()
            cached_repo_hash = self._get_cache_hash()
            if cached_repo_hash == repo_hash:
                self.hash = repo_hash
                self.data = yaml.load(open(self.cache_path),
                                      Loader=YLoader)
                logger.info("Load data from the cache.")
            else:
                logger.info("DB cache is outdated.")

    def _update_git_clone(self):
        repo = git.Git(self.clone_path)
        repo.init()
        try:
            repo.execute(['git', 'remote', 'add',
                          'origin', self.git_repo_url])
        except Exception:
            logger.info("Re-using the previous repo path %s" % self.clone_path)
            repo.execute(['git', 'remote', 'remove', 'origin'])
            repo.execute(['git', 'remote', 'add',
                          'origin', self.git_repo_url])
            logger.info("Update the previous remote origin to %s." % (
                        self.git_repo_url))
        repo.execute(['git', 'fetch', 'origin'])
        repo.execute(['git', 'checkout', '-B', 'master', 'origin/master'])
        repo.execute(['git', 'reset', '--hard', self.git_ref])
        logger.info("Updated GIT repo %s from %s at ref %s." % (
            self.clone_path, self.git_repo_url, self.git_ref))

    def _load_db(self):
        def check_ext(f):
            return f.endswith('.yaml') or f.endswith('.yml')
        logger.info("Load data from the YAML files.")
        self.rids = {}
        yamlfiles = [f for f in os.listdir(self.db_path) if check_ext(f)]
        for f in yamlfiles:
            logger.info("Reading %s ..." % f)
            try:
                yaml_data = yaml.load(
                    open(os.path.join(self.db_path, f)),
                    Loader=YLoader)
            except Exception:
                raise YAMLDBException(
                    "YAML format corrupted in file %s" % (
                        os.path.join(self.db_path, f)))
            if not self.data:
                self.data = self.validate(yaml_data, self.rids)
            else:
                data_to_append = self.validate(yaml_data, self.rids)
                for rtype, resources in data_to_append['resources'].items():
                    if rtype not in self.data['resources']:
                        self.data['resources'][rtype] = {}
                    self.data['resources'][rtype].update(resources)

    @staticmethod
    def _validate_base_struct(data):
        try:
            assert isinstance(data, type(RESOURCES_STRUCT))
            assert isinstance(data['resources'],
                              type(RESOURCES_STRUCT['resources']))
        except (AssertionError, KeyError):
            raise YAMLDBException(
                "The main resource data structure is invalid")
        try:
            for rtype, resources in data['resources'].items():
                assert isinstance(
                    rtype, type(list(RESOURCES_STRUCT['resources'].keys())[0]))
                assert isinstance(
                    resources, type(RESOURCES_STRUCT['resources']['rtype']))
        except AssertionError:
            raise YAMLDBException(
                "Resource type %s structure is invalid" % rtype)
        try:
            for rtype, resources in data['resources'].items():
                for rid, resource in resources.items():
                    assert isinstance(rid, str)
                    assert isinstance(
                        resource,
                        type(RESOURCES_STRUCT['resources']['rtype']['key']))
        except AssertionError:
            raise YAMLDBException(
                "Resource %s of type %s is invalid" % (resource, rtype))

    @staticmethod
    def _validate_rid_unicity(data, rids):
        # Verify at YAML load time that duplicated resources key
        # are not present. To avoid overlapping of resources.
        # https://gist.github.com/pypt/94d747fe5180851196eb implements a
        # solution but seems difficult as it does not support the
        # safe_loader and usage of that loader is important to avoid
        # loading malicious yaml serialized objects.
        # Inside a single yaml file a reviewer should take care to
        # the duplicated keys issue and also look at the logs for
        # checking not expected detected changes.
        #
        # Nevertheless between two or more yaml files this check can be
        # implemented.
        for rtype, resources in data['resources'].items():
            for rid, resource in resources.items():
                rids.setdefault(rtype, {})
                if rid not in rids[rtype]:
                    rids[rtype][rid] = None
                else:
                    raise YAMLDBException(
                        "Duplicated resource ID detected for "
                        "resource type: %s id: %s" % (rtype, rid))

    def refresh(self):
        """ Reload of the YAML files.
        """
        self.data = None
        self._update_git_clone()
        self._load_from_cache()
        # Load from files. Cache is not up to date.
        if not self.data:
            self._load_db()
            self._update_cache()

    @staticmethod
    def validate(data, rids):
        """ Validate the resource data structure.
        """
        YAMLBackend._validate_base_struct(data)
        YAMLBackend._validate_rid_unicity(data, rids)
        return data

    def get_data(self):
        """ Return the full data structure.
        """
        data = self.data
        data['hash'] = self.hash
        return data


class MemoryYAMLBackend(YAMLBackend):
    def __init__(self, buffer_d):
        """ Class to read and validate resources from a dict
        of buffers. Buffers should each be a YAML content.
        keys are files name, and values YAML content.

        :param data: a dict of buffer
        """
        super(MemoryYAMLBackend, self)
        self.data = None
        self.buffer_d = buffer_d
        self.refresh()

    def _load_db(self):
        logger.info("Load data from the YAMLs buffer.")
        self.rids = {}
        for k, v in self.buffer_d.items():
            logger.info("Reading buffer %s ..." % k)
            try:
                if isinstance(v, dict):
                    yaml_data = v
                else:
                    yaml_data = yaml.load(
                        StringIO(v),
                        Loader=YLoader)
            except Exception:
                raise YAMLDBException(
                    "YAML format corrupted in buffer %s" % k)
            if not self.data:
                self.data = self.validate(yaml_data, self.rids)
            else:
                data_to_append = self.validate(yaml_data, self.rids)
                for rtype, resources in data_to_append['resources'].items():
                    if rtype not in self.data['resources']:
                        self.data['resources'][rtype] = {}
                    self.data['resources'][rtype].update(resources)

    def refresh(self):
        """ Reload of the YAML buffers.
        """
        self.data = None
        if not self.data:
            self._load_db()

    def get_data(self):
        """ Return the full data structure.
        """
        return self.data



# Copyright (C) 2015  Red Hat <licensing@enovance.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import TestCase
from mock import patch

from managesf.tests import dummy_conf
from managesf.services import gerrit


class BaseSFGerritService(TestCase):
    @classmethod
    def setupClass(cls):
        cls.conf = dummy_conf()
        cls.gerrit = gerrit.SoftwareFactoryGerrit(cls.conf)


class TestSFGerritUserManager(BaseSFGerritService):
    user_data = {"_account_id": 5,
                 "name": "Jotaro Kujoh",
                 "email": "jojo@starplatinum.dom",
                 "username": "jojo",
                 "avatars": [{"url": "meh",
                              "height": 26}]}

    def test_update(self):
        # TODO: implement it
        pass

    def test_create(self):
        with patch.object(self.gerrit.user,
                          '_add_account_as_external') as add_external, \
                    patch('managesf.services.gerrit.utils.GerritClient.'
                          'create_account') as create:
            create.return_value = self.user_data
            self.gerrit.user.create('jojo', 'jojo@starplatinum.dom',
                                    'Jotaro Kujoh', None, 10)
            _user = {'name': str('Jotaro Kujoh'),
                     'email': str('jojo@starplatinum.dom')}
            create.assert_called_with('jojo', _user)
            add_external.assert_called_with(5, 'jojo')
            create.reset_mock()
            _user["ssh_key"] = 'bop'
            self.gerrit.user.create('jojo', 'jojo@starplatinum.dom',
                                    'Jotaro Kujoh', [{'key': 'bop'}], 10)
            create.assert_called_with('jojo', _user)
        # Test fringe case where we try to create the admin user
        self.assertEqual(1,
                         self.gerrit.user.create(None, 'dio@wryyyyy.org',
                                                 'Dio Brando', None, 1))
        # TODO: Add case where the user exists and update is called by create

    def test_get(self):
        with patch('managesf.services.gerrit.utils.GerritClient.'
                   'get_account') as get:
            get.return_value = self.user_data
            u = self.gerrit.user.get('jojo')
            get.assert_called_with('jojo')
            self.assertEqual(5, u)

    def test_delete(self):
        self.assertEquals(self.gerrit.user.delete('jojo'), None)



#
# Copyright (C) 2017 Red Hat
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import traceback

from pecan import expose
from pecan import request, response, abort

from managesf.controllers.api.v2 import base
from managesf.api.v2.managers import resource_manager as manager


class ResourcesRootController(base.APIv2RestController):

    @expose('json')
    def get(self, **kwargs):
        # TODO change this when group lookup is fixed
        _policy = 'managesf.resources:get'
        if not kwargs:
            kwargs = request.json if request.content_length else {}
        if kwargs.get('get_missing_resources', '').lower() == 'true':
            kwargs['get_missing_resources'] = True
        else:
            kwargs['get_missing_resources'] = False
        target = dict((k, v) for k, v in kwargs.items()
                      if k not in ['order_by', 'skip', 'limit'])
        if not base.authorize(_policy,
                              target=target):
            return abort(401,
                         detail='Failure to comply with policy %s' % _policy)
        response.status = 400
        try:
            results = manager.resources.get(**kwargs)
            response.status = 200
            return results
        except NotImplementedError as e:
            response.status = 404
            return {'error_description': str(e)}
        except Exception as e:
            response.status = 500
            self._logger.exception(e)
            return {'error_description': str(e),
                    'traceback': traceback.format_exc()}

    @expose('json')
    def post(self, **kwargs):
        _policy = 'managesf.resources:validate'
        if not kwargs:
            kwargs = request.json if request.content_length else {}
        if not base.authorize(_policy,
                              target=kwargs):
            return abort(401,
                         detail='Failure to comply with policy %s' % _policy)
        try:
            status, logs = manager.resources.create(**kwargs)
            if not status:
                response.status = 409
            else:
                response.status = 200
            return logs
        except ValueError as e:
            response.status = 400
            return {'error_description': str(e)}
        except NotImplementedError as e:
            response.status = 404
            return {'error_description': str(e)}
        except Exception as e:
            response.status = 500
            self._logger.exception(e)
            return {'error_description': str(e)}



#
# Copyright (c) 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import re


class ModelInvalidException(Exception):
    pass


class ResourceInvalidException(Exception):
    pass


AUTHORIZED_CALLBACKS = ('update', 'create', 'delete',
                        'extra_validations', 'get_all')

KEY_RE_CONSTRAINT = "[a-zA-Z0-9-_]+"


class BaseResource(object):
    """ All resources to define for this backend
    must inherit from this class. This class
    cannot be initialized by itself.

    The resource data model description must
    contains a dictionnary where each key is a
    resource attribute and value a tuple that describes
    the attribute constraints.

    'key': (
        str, # Type
        regexp, # Value regex validator
        True, # Mandatory key
        None, # Default value
        True, # Is the value mutable
        "String", # Description
    ),

    """

    MODEL_TYPE = 'default'
    DESCRIPTION = 'No description available'
    MODEL = {}
    PRIORITY = None
    # The primary key tells if a field should be unique
    # over all resources of a given type. Leave it at None if not the case
    PRIMARY_KEY = None
    CALLBACKS = {
        'update': lambda conf, new, kwargs: NotImplementedError,
        'create': lambda conf, new, kwargs: NotImplementedError,
        'delete': lambda conf, new, kwargs: NotImplementedError,
        'extra_validations': lambda conf, new, kwargs: NotImplementedError,
        'get_all': lambda conf, new: NotImplementedError,
    }

    def __init__(self, id, resource):
        self.id = id
        self._model_definition_validate()
        self.resource = resource
        # "name" is a special key that will inherit of the Resource
        # ID value. The ID value must then match the "name" regex
        # contraint
        self.resource['name'] = self.id
        self.mandatory_keys = set(
            [k for k, v in self.__class__.MODEL.items() if v[2]])
        self.keys = set(self.__class__.MODEL)

    def _model_definition_validate(self):
        """ This validate the inherited model. This is
        to validate resource model defined by inherited
        classes. We make sure the model is followed by the
        developper.
        """
        try:
            assert isinstance(self.__class__.MODEL_TYPE, str)
            assert (isinstance(self.__class__.PRIMARY_KEY, str) or
                    self.__class__.PRIMARY_KEY is None)
            assert isinstance(self.__class__.PRIORITY, int)
            for key in self.__class__.MODEL:
                assert re.match("^" + KEY_RE_CONSTRAINT + "$", key)
            assert 'name' in self.__class__.MODEL
        except Exception:
            raise ModelInvalidException(
                "Model %s is invalid and not usable" % (
                    self.__class__.MODEL_TYPE))

        if self.__class__.PRIMARY_KEY and self.__class__.PRIMARY_KEY != 'name':
            if self.__class__.PRIMARY_KEY not in self.__class__.MODEL:
                raise ModelInvalidException(
                    "Model %s primary key %s does not exists" % (
                        self.__class__.MODEL_TYPE,
                        self.__class__.PRIMARY_KEY))

            if not self.__class__.MODEL[self.__class__.PRIMARY_KEY][2]:
                raise ModelInvalidException(
                    "Model %s primary key %s should be mandatory" % (
                        self.__class__.MODEL_TYPE,
                        self.__class__.PRIMARY_KEY))

        for constraints in self.__class__.MODEL.values():
            if len(constraints) != 6:
                raise ModelInvalidException(
                    "Model %s is invalid and not usable "
                    "(missing field)" % (
                        self.__class__.MODEL_TYPE))

        try:
            # Be sure default values are of the declared type
            # make some others validation on default value
            for key, constraints in self.__class__.MODEL.items():
                # Only act on non-mandatory keys as default
                # is provided. Skip 'name' checking.
                if not constraints[2] and key != 'name':
                    # Validate default value type
                    assert isinstance(constraints[3],
                                      constraints[0])
                    # Validate default value match the regexp
                    # if str type
                    if constraints[0] is str:
                        assert re.match(constraints[1],
                                        constraints[3])
                    # Validate default value match the regexp
                    # if dict type
                    if constraints[0] is dict:
                        assert isinstance(constraints[1], tuple)
                        key_re = re.compile(constraints[1][0])
                        val_re = re.compile(constraints[1][1])
                        for k, v in constraints[3].items():
                            assert key_re.match(k)
                            if isinstance(v, (str, bytes)):
                                assert val_re.match(v)
                            else:
                                if (not isinstance(v, bool) and
                                        not isinstance(v, int)):
                                    raise AssertionError()
                    # Validate list default values match the regexp
                    # if list type
                    if constraints[0] is list:
                        # Constraints[1] is string (a regexp) in case of list
                        # of str only
                        if isinstance(constraints[1], (str, bytes)):
                            assert all([re.match(constraints[1], c) for
                                        c in constraints[3]]) is True
                        # If constraints[1] is a tuple then constraints[1][0]
                        # must be a dict and constraints[1][1] is regexp for
                        # the dict key. But value can still be a simple string
                        # then the value is validated against constraints[1][1]
                        elif isinstance(constraints[1], tuple):
                            assert constraints[1][0] is dict
                            assert isinstance(constraints[1][1], str)
                            for c in constraints[3]:
                                if isinstance(c, dict):
                                    # Make sure default value only have dict
                                    # with a single key
                                    assert len(c.keys()) == 1
                                    # Make sure default value only have dict
                                    # with a key matching the regexp
                                    assert re.match(
                                        constraints[1][1], list(c)[0])
                                if isinstance(c, (str, bytes)):
                                    assert re.match(constraints[1][1], c)
                        else:
                            # Unsuported case constraints[1] can be only a
                            # (str, bytes) or tuple
                            assert False
        except Exception:
            raise ModelInvalidException(
                "Model %s is invalid and not usable "
                "(Wrong default value according to the type "
                "or regex)" % (
                    self.__class__.MODEL_TYPE))

        # Validate the callbacks of the inherited model
        try:
            # Be sure we have only the authorized callbacks
            assert len(set(AUTHORIZED_CALLBACKS).symmetric_difference(
                set(self.__class__.CALLBACKS))) == 0
            # Be sure the callbacks are callable or NotImplemented
            for key, callback in self.__class__.CALLBACKS.items():
                if (not callable(callback)
                        and callback is not NotImplementedError):
                    raise Exception
        except Exception:
            raise ModelInvalidException(
                "Model %s callbacks are invalid, model is not usable" % (
                    self.__class__.MODEL_TYPE))

    def val_validate(self, v, val_re):
        return val_re.match(v) if isinstance(v, (str, bytes)) else True

    def validate(self):
        """ Validate the data MODEL of the resource
        """
        # Validate all mandatory keys are present
        if not self.mandatory_keys.issubset(set(self.resource)):
            raise ResourceInvalidException(
                "Resource [type: %s, ID: %s] miss a "
                "mandatory key. Please check the model." % (
                    self.__class__.MODEL_TYPE,
                    self.id))

        # Validate the resource does not contains extra keys
        if not set(self.resource).issubset(self.keys):
            raise ResourceInvalidException(
                "Resource [type: %s, ID: %s] contains "
                "extra keys. Please check the model." % (
                    self.__class__.MODEL_TYPE,
                    self.id))

        # Validate the resource value type
        for key, value in self.resource.items():
            if not isinstance(value, self.__class__.MODEL[key][0]):
                raise ResourceInvalidException(
                    "Resource [type: %s, ID: %s] has an invalid "
                    "key (%s) data type (expected: %s)" % (
                        self.__class__.MODEL_TYPE,
                        self.id,
                        key,
                        self.__class__.MODEL[key][0]))
            # For str type validate the content as according the regex
            if self.__class__.MODEL[key][0] is str:
                if not re.match(self.__class__.MODEL[key][1], value):
                    raise ResourceInvalidException(
                        "Resource [type: %s, ID: %s] has an invalid "
                        "key (%s) data content (expected match : %s)" % (
                            self.__class__.MODEL_TYPE,
                            self.id,
                            key,
                            self.__class__.MODEL[key][1]))
            # For list type validate the content as according the regex, we
            # expect a string if MODEL[key][1] is a string. But the list will
            # contain dictionaries or string if MODEL[key][1] is a tuple
            if self.__class__.MODEL[key][0] is list:
                if isinstance(self.__class__.MODEL[key][1], (str, bytes)):
                    # List values can be only string
                    if not all([isinstance(v, (str, bytes)) for v in value]):
                        raise ResourceInvalidException(
                            "Resource [type: %s, ID: %s] has an "
                            "invalid key (%s) data content (expected "
                            "a string)" % (
                                self.__class__.MODEL_TYPE,
                                self.id,
                                key))
                    if not all([re.match(self.__class__.MODEL[key][1], v)
                                for v in value]):
                        raise ResourceInvalidException(
                            "Resource [type: %s, ID: %s] has an "
                            "invalid key (%s) data content (expected "
                            "match : %s)" % (
                                self.__class__.MODEL_TYPE,
                                self.id,
                                key,
                                self.__class__.MODEL[key][1]))
                else:
                    # We assume self.__class__.MODEL[key][1] is tuple. List
                    # values can be dict or list
                    for v in value:
                        if isinstance(v, (str, bytes)):
                            if not re.match(
                                    self.__class__.MODEL[key][1][1], v):
                                raise ResourceInvalidException(
                                    "Resource [type: %s, ID: %s] has an "
                                    "invalid key (%s) data content (expected "
                                    "match : %s)" % (
                                        self.__class__.MODEL_TYPE,
                                        self.id,
                                        key,
                                        self.__class__.MODEL[key][1][1]))
                        elif isinstance(v, dict):
                            if not len(v.keys()) == 1:
                                raise ResourceInvalidException(
                                    "Resource [type: %s, ID: %s] has an "
                                    "invalid key (%s) data content. List "
                                    "contains a dictionary with multiple "
                                    "keys" % (
                                        self.__class__.MODEL_TYPE,
                                        self.id,
                                        key))
                            if not re.match(
                                    self.__class__.MODEL[key][1][1],
                                    list(v.keys())[0]):
                                raise ResourceInvalidException(
                                    "Resource [type: %s, ID: %s] has an "
                                    "invalid key (%s) data content (expected "
                                    "match : %s)" % (
                                        self.__class__.MODEL_TYPE,
                                        self.id,
                                        key,
                                        self.__class__.MODEL[key][1]))
                        else:
                            raise ResourceInvalidException(
                                "Resource [type: %s, ID: %s] key %s contains "
                                "unsuported data type. Please check the "
                                " model." % (self.__class__.MODEL_TYPE,
                                             self.id, key))
            # For dict type validate the content as according the regex
            if self.__class__.MODEL[key][0] is dict:
                try:
                    for k, v in value.items():
                        assert isinstance(k, str)
                        try:
                            assert isinstance(v, str)
                        except Exception:
                            if (not isinstance(v, bool) and
                                    not isinstance(v, int)):
                                raise AssertionError()
                except Exception:
                    raise ResourceInvalidException(
                        "Resource [type: %s, ID: %s] has an invalid "
                        "key (%s) dict keys or values not valid" % (
                            self.__class__.MODEL_TYPE,
                            self.id,
                            key))
                key_re = re.compile(self.__class__.MODEL[key][1][0])
                val_re = re.compile(self.__class__.MODEL[key][1][1])
                if not all([all([key_re.match(k),
                                self.val_validate(v, val_re)]) for
                            k, v in value.items()]):
                    raise ResourceInvalidException(
                        "Resource [type: %s, ID: %s] has an invalid "
                        "key (%s) dict key, value (expected match : %s)" % (
                            self.__class__.MODEL_TYPE,
                            self.id,
                            key,
                            self.__class__.MODEL[key][1]))

    def is_mutable(self, key):
        return self.__class__.MODEL[key][4]

    def get_deps(self, keyname=False):
        """ Return a dictionnary of {rtype: set([ids, ])} that list
        resource instance dependencies. eg. a given
        resource field requires another resource to be validated, or
        applied. Note that dependencies between resources of the
        same type cannot exists.

        If keyname is set to True then this method should return
        the key name of the resource that contains dependency ids.
        """
        if keyname:
            return ''
        return {}

    def is_deps_soft(self):
        """ If this returns True contraints dependencies check become
        soft. That means resources ids returned by get_deps won't be
        checked for existence.
        """
        return False

    def set_defaults(self, soft=False):
        """ Enrich the data MODEL. This method add
        missing fields to the resource. Missing fields are
        initialized with their default value.
        If soft is True then only default value that is data (not [] or '')
        will be added as default value.
        """
        for key, constraints in self.__class__.MODEL.items():
            if key not in self.resource:
                if soft and not constraints[3]:
                    continue
                self.resource[key] = constraints[3]

    def get_resource(self):
        return self.resource

    def should_be_updated(self):
        """ Is the resource should be updated by inheritance
        if one of the resources it depends on (only one
        rtype possible) has been updated.
        """
        return True

    def transform_for_get(self):
        """ This method can be overwritten to format of the resource
        when a get on the resource tree is asked.
        """
        return self.resource



# Copyright (C) 2014 eNovance SAS <licensing@enovance.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import TestCase
from mock import patch, call

import os

from managesf.services.gerrit import utils
from managesf.tests import dummy_conf


class TestGerritRepo(TestCase):
    @classmethod
    def setupClass(cls):
        cls.conf = dummy_conf()

    def test_init(self):
        gr = utils.GerritRepo('p1', self.conf)
        self.assertTrue(gr.infos['localcopy_path'].endswith('p1'))
        self.assertTrue(os.path.isfile(gr.env['GIT_SSH']))
        self.assertTrue(gr.env['GIT_COMMITTER_NAME'])
        self.assertTrue(gr.env['GIT_COMMITTER_EMAIL'])

    def test_exec(self):
        gr = utils.GerritRepo('p1', self.conf)
        gr._exec('touch f')
        self.assertTrue(os.path.isfile(os.path.join(gr.infos['localcopy_path'],
                                                    'f')))

    def test_clone(self):
        gr = utils.GerritRepo('p1', self.conf)
        with patch.object(gr, '_exec') as ex:
            gr.clone()
            self.assertEqual(
                'git clone ssh://user1@gerrit.test.dom:2929/p1 %s' %
                gr.infos['localcopy_path'],
                ex.mock_calls[0][1][0])

    def test_add_file(self):
        gr = utils.GerritRepo('p1', self.conf)
        with patch.object(gr, '_exec') as ex:
            gr.add_file('thefile', 'thecontent')
            p = os.path.join(gr.infos['localcopy_path'], 'thefile')
            self.assertTrue(os.path.isfile(p))
            self.assertEqual('thecontent', open(p).read())
            self.assertEqual('git add thefile', ex.mock_calls[0][1][0])

    def test_push_config(self):
        gr = utils.GerritRepo('p1', self.conf)
        with patch.object(gr, '_exec') as ex:
            with patch.object(gr, 'add_file') as af:
                gr.push_config({'f1': 'contentf1', 'f2': 'contentf2'})
                self.assertEqual(2, len(af.mock_calls))
                self.assertEqual(6, len(ex.mock_calls))

    def test_push_branch(self):
        gr = utils.GerritRepo('p1', self.conf)
        with patch.object(gr, '_exec') as ex:
            with patch.object(gr, 'add_file') as af:
                ex.return_value = True
                gr.push_branch('master', {'f1': 'contentf1',
                                          'f2': 'contentf2'})
        self.assertListEqual(
            [call('git checkout master'),
             call('git reset --hard origin/master'),
             call('git status -s'),
             call("git commit -a --author 'user1 "
                  "<user1@tests.dom>' -m'ManageSF commit'"),
             call('git push origin master')],
            ex.call_args_list)
        self.assertTrue(len(af.call_args_list), 2)

    def test_push_master_from_git_remote(self):
        gr = utils.GerritRepo('p1', self.conf)
        with patch.object(gr, '_exec') as ex:
            gr.push_master_from_git_remote('git://tests.dom/git/oldp1.git')
            self.assertEqual(5, len(ex.mock_calls))

    def test_review_changes(self):
        gr = utils.GerritRepo('p1', self.conf)
        with patch.object(gr, '_exec') as ex:
            gr.review_changes('this is a test')
            self.assertEqual(3, len(ex.mock_calls))
            self.assertTrue(ex.mock_calls[0][1][0].startswith(
                'ssh-agent bash -c'))
            self.assertTrue(ex.mock_calls[1][1][0].startswith('git commit -a'))
            self.assertEqual('git review', ex.mock_calls[2][1][0])

    def test_list_remote_branches(self):
        gr = utils.GerritRepo('p1', self.conf)
        with patch.object(gr, '_exec') as ex:
            ex.return_value = """  origin/HEAD   -> origin/master
  origin/master 9dc37aee187412073a10c9df85b6878bc39bd1a2 Cmt msg
  origin/meta/config 3d5c40c888109a69c1211281a67c6dbaadf7ae56 Provides ACLs
  origin/stable/ocata 4d5c40c888109a69c1211281a67c6dbaadf7ae56 ABC
"""
            refs = gr.list_remote_branches()
            self.assertDictEqual(
                refs,
                {'HEAD': 'master',
                 'master': '9dc37aee187412073a10c9df85b6878bc39bd1a2',
                 'stable/ocata': '4d5c40c888109a69c1211281a67c6dbaadf7ae56',
                 'meta/config': '3d5c40c888109a69c1211281a67c6dbaadf7ae56'})

        with patch.object(gr, '_exec') as ex:
            ex.return_value = """  origin/HEAD   -> origin/stable/ocata
  origin/master 9dc37aee187412073a10c9df85b6878bc39bd1a2 Cmt msg
  origin/meta/config 3d5c40c888109a69c1211281a67c6dbaadf7ae56 Provides ACLs
  origin/stable/ocata 4d5c40c888109a69c1211281a67c6dbaadf7ae56 ABC
"""
            refs = gr.list_remote_branches()
            self.assertDictEqual(
                refs,
                {'HEAD': 'stable/ocata',
                 'master': '9dc37aee187412073a10c9df85b6878bc39bd1a2',
                 'stable/ocata': '4d5c40c888109a69c1211281a67c6dbaadf7ae56',
                 'meta/config': '3d5c40c888109a69c1211281a67c6dbaadf7ae56'})

    def test_create_remote_branch(self):
        gr = utils.GerritRepo('p1', self.conf)
        with patch.object(gr, '_exec') as ex:
            with patch.object(gr, 'list_remote_branches') as lrb:
                lrb.return_value = {}
                gr.create_remote_branch('mybranch', '123')
                self.assertEqual(
                    'git branch mybranch 123',
                    ex.mock_calls[0][1][0])
                self.assertEqual(
                    'git push origin mybranch',
                    ex.mock_calls[1][1][0])
        with patch.object(gr, '_exec') as ex:
            with patch.object(gr, 'list_remote_branches') as lrb:
                lrb.return_value = {'fromthisbranch': '124'}
                gr.create_remote_branch('mybranch', 'fromthisbranch')
                self.assertEqual(
                    'git branch mybranch 124',
                    ex.mock_calls[0][1][0])
                self.assertEqual(
                    'git push origin mybranch',
                    ex.mock_calls[1][1][0])

    def test_delete_remote_branch(self):
        gr = utils.GerritRepo('p1', self.conf)
        with patch.object(gr, '_exec') as ex:
            gr.delete_remote_branch('mybranch')
            self.assertEqual(
                'git push --delete origin mybranch',
                ex.mock_calls[0][1][0])



#
# Copyright (C) 2017 Red Hat
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import logging

from managesf.api.v2 import base


logger = logging.getLogger(__name__)


resource_manager = None
logger.info('Loading resources manager...')
try:
    resource_manager = base.load_manager('managesf.v2.resources',
                                         'manageSF')
except Exception as e:
    msg = 'Cannot load resource service: %s' % e
    logger.error(msg)
    raise



#
# Copyright (C) 2017 Red Hat
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from managesf.api.v2 import base


class ResourcesManager(base.BaseCRUDManager):
    """Resource tree-related CRUD operations."""

    def __init__(self):
        super(ResourcesManager, self).__init__()

    def get(self, **kwargs):
        """get the full tree.
        get_missing_resources: (boolean) if True, returns existing projects
            that are missing from the resources descriptor.
        """
        raise NotImplementedError

    def create(self, **kwargs):
        raise NotImplementedError


class ResourcesServiceManager(base.BaseService):
    resources = ResourcesManager()
